import time
import random
import sys
import os
import weakref
from box import Box
from anytree import PreOrderIter

import opengate_core as g4
from .exception import fatal, warning, GateImplementationError
from .decorators import requires_fatal, requires_warning
from .runtiming import assert_run_timing
from .uisessions import UIsessionSilent, UIsessionVerbose
from .exception import ExceptionHandler
from .physics import (
    UserLimitsPhysics,
    translate_particle_name_gate_to_geant4,
    cut_particle_names,
    create_g4_optical_properties_table,
    load_optical_properties_from_xml,
)
from .base import GateSingletonFatal
from .logger import logger


class EngineBase:
    """
    Base class for all engines (SimulationEngine, VolumeEngine, etc.)
    """

    def __init__(self, simulation_engine) -> None:
        self.simulation_engine = simulation_engine
        # debug verbose
        self.verbose_getstate = simulation_engine.verbose_getstate
        self.verbose_close = simulation_engine.verbose_close

    def close(self):
        self.simulation_engine = None

    def __getstate__(self):
        raise GateImplementationError(
            f"The __getstate__() method of the {type(self).__name__} class got called. "
            f"That should not happen because it should be closed before anything is pickled "
            f"at the end of a subprocess. Check warning messages for clues!"
        )


class SourceEngine(EngineBase):
    """
    Source Engine manages the G4 objects of sources at runtime
    """

    # G4RunManager::BeamOn takes an int as input. The max cpp int value is currently 2147483647
    # Python manages int differently (no limit), so we need to set the max value here.
    max_int = 2147483647

    def __init__(self, simulation_engine):
        super().__init__(simulation_engine)

        # Keep a pointer to the current simulation
        # self.source_manager = source_manager
        self.simulation_engine = simulation_engine

        # List of run time intervals
        self.run_timing_intervals = None
        self.current_run_interval = None

        # use a progress bar ?
        self.progress_bar = False
        self.expected_number_of_events = "unknown"

        # List of sources (GateSource), for all threads
        self.sources = []

        # The source manager will be constructed at build (during ActionManager)
        # Its task is to call GeneratePrimaries and loop over the sources
        # For MT, the master_source_manager is the MasterThread
        # The g4_thread_source_managers list all master sources for all threads
        self.g4_master_source_manager = None
        self.g4_thread_source_managers = []

        # Options dict for cpp SourceManager
        # will be set in create_g4_source_manager
        # FIXME: Why is this separate dictionary needed? Would be better to access the source manager directly
        self.source_manager_options = Box()

    def close(self):
        if self.verbose_close:
            warning("Closing SourceEngine")
        self.release_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_master_source_manager = None
        self.g4_thread_source_managers = None
        # a source object contains a reference to a G4 source
        self.sources = None

    def initialize(self, run_timing_intervals, progress_bar=False):
        self.run_timing_intervals = run_timing_intervals
        assert_run_timing(self.run_timing_intervals)
        # if len(self.simulation_engine.simulation.source_manager.sources) == 0:
        #    self.simulation_engine.simulation.warn_user(
        #        "No source: no particle will be generated"
        #    )
        self.progress_bar = progress_bar

    def initialize_actors(self):
        """
        Set all actors to the master source manager
        """
        self.g4_master_source_manager.SetActors(
            self.simulation_engine.simulation.actor_manager.sorted_actors
        )

    def create_master_source_manager(self):
        # create the master source for the masterThread
        self.g4_master_source_manager = self.create_g4_thread_source_manager(
            append=False
        )
        return self.g4_master_source_manager

    def create_g4_thread_source_manager(self, append=True):
        """
        This is called by all threads.
        This object is needed here because it can only be
        created after physics initialisation
        """
        # create a source manager for the current thread
        ms = g4.GateSourceManager()
        # create all sources for this source manager (for all threads)
        source_manager = self.simulation_engine.simulation.source_manager
        for source in source_manager.sources.values():
            source.initialize(self.run_timing_intervals)
            source.add_to_source_manager(ms)
            # store all the sources (will be used later by SimulationOutput)
            self.sources.append(source)

        # Copy visualization parameters
        for k, v in self.simulation_engine.simulation.user_info.items():
            if "visu" in k:
                self.source_manager_options[k] = v

        # Copy verbosity parameters
        for k, v in self.simulation_engine.simulation.user_info.items():
            if "verbose_" in k:
                self.source_manager_options[k] = v

        # progress bar
        self.source_manager_options["progress_bar"] = (
            self.simulation_engine.simulation.progress_bar
        )

        ms.Initialize(self.run_timing_intervals, self.source_manager_options)
        self.expected_number_of_events = (
            ms.GetExpectedNumberOfEvents()
            * self.simulation_engine.simulation.number_of_threads
        )
        # set the flag for user event info
        ms.fUserEventInformationFlag = (
            self.simulation_engine.user_event_information_flag
        )
        # keep the pointer to avoid deletion
        if append:
            self.g4_thread_source_managers.append(ms)

        return ms

    def start(self):
        # FIXME (1) later : may replace BeamOn with DoEventLoop
        # to allow better control on geometry between the different runs
        # FIXME (2) : check estimated nb of particle, warning if too large
        # start the master thread (only main thread)
        for (
            solid
        ) in self.simulation_engine.simulation.volume_manager.solid_with_texture_init:
            self.g4_master_source_manager.RegisterImageBox(solid)
        self.simulation_engine.simulation.volume_manager.solid_with_texture_init = []

        self.g4_master_source_manager.StartMasterThread()

        # once terminated, packup the sources (if needed)
        for source in self.sources:
            source.prepare_output()

    def can_predict_expected_number_of_event(self):
        # can_predict = True
        # for source in self.sources:
        #     can_predict = can_predict and source.can_predict_number_of_events()
        return all(s.can_predict_number_of_events() for s in self.sources)


class PhysicsEngine(EngineBase):
    """
    Class that contains all the information and mechanism regarding physics
    to actually run a simulation. It is associated with a simulation engine.

    """

    def __init__(self, *args):
        super().__init__(*args)
        # Keep a shortcut reference to the current physics_manager
        self.physics_manager = self.simulation_engine.simulation.physics_manager

        # Register this engine with the regions
        for region in self.physics_manager.regions.values():
            region.physics_engine = self

        for optical_surface in self.physics_manager.optical_surfaces.values():
            optical_surface.physics_engine = self

        # main g4 physic list
        self.g4_physics_list = None
        self.g4_decay = None
        self.g4_radioactive_decay = None
        self.g4_cuts_by_regions = []
        self.g4_em_parameters = None
        self.g4_parallel_world_physics = []
        self.g4_optical_material_tables = {}
        self.g4_physical_volumes = []
        self.g4_surface_properties = None

        # We need to keep a reference to physics constructors
        # implemented on the python side
        # Physics constructor linked via pybind have the nodelete pointer option,
        # so python will not delete them and no reference needs to be kept
        self.gate_physics_constructors = []

        self.optical_surfaces_properties_dict = {}

    def close(self):
        if self.verbose_close:
            warning("Closing PhysicsEngine")
        self.close_physics_constructors()
        self.release_g4_references()
        self.release_optical_surface_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_physics_list = None
        self.g4_decay = None
        self.g4_radioactive_decay = None
        self.g4_cuts_by_regions = None
        self.g4_em_parameters = None
        self.g4_parallel_world_physics = []
        self.g4_optical_material_tables = {}
        self.g4_physical_volumes = []
        self.g4_surface_properties = None

    def release_optical_surface_g4_references(self):
        for optical_surface in self.physics_manager.optical_surfaces.values():
            optical_surface.release_g4_references()

    @requires_fatal("simulation_engine")
    @requires_warning("g4_physics_list")
    def close_physics_constructors(self):
        """This method removes PhysicsConstructors defined in python from the physics list.

        It should be called after a simulation run, i.e. when a simulation engine closes,
        because the RunManager will otherwise attempt to delete the PhysicsConstructor
        and cause a segfault.

        """
        current_state = self.simulation_engine.g4_state
        self.simulation_engine.g4_state = g4.G4ApplicationState.G4State_PreInit
        for pc in self.gate_physics_constructors:
            self.g4_physics_list.RemovePhysics(pc)
        self.simulation_engine.g4_state = current_state

    # make this a property so the communication between
    # PhysicsManager and PhysicsEngine can be changed without
    # impacting this class
    @property
    def user_info_physics_manager(self):
        return self.physics_manager.user_info

    def initialize_before_runmanager(self):
        """Initialize methods to be called *before*
        G4RunManager.Initialize() is called.

        """
        self.initialize_physics_list()
        self.initialize_g4_em_parameters()
        self.initialize_user_limits_physics()
        self.initialize_physics_biasing()
        self.initialize_parallel_world_physics()

    def initialize_after_runmanager(self):
        """ """
        # Cuts need to be set *after*
        # G4RunManager.Initialize() is called.
        # Reason: The Initialize() sequence would otherwise override
        # the global cuts with the physics list defaults.
        self.initialize_global_cuts()
        self.initialize_regions()
        self.initialize_optical_material_properties()
        self.initialize_optical_surfaces()
        self.initialize_ionisation_options()
        self.initialize_users_ionisation_potentials()

    def initialize_parallel_world_physics(self):
        for (
            world
        ) in self.physics_manager.simulation.volume_manager.parallel_world_names:
            pwp = g4.G4ParallelWorldPhysics(world, True)
            self.g4_parallel_world_physics.append(pwp)
            self.g4_physics_list.RegisterPhysics(pwp)

    def initialize_physics_list(self):
        """
        Create a Physic List from the Factory
        """
        physics_list_name = self.physics_manager.user_info.physics_list_name
        self.g4_physics_list = (
            self.physics_manager.physics_list_manager.get_physics_list(
                physics_list_name
            )
        )

    def initialize_regions(self):
        for region in self.physics_manager.regions.values():
            region.initialize()

    def initialize_global_cuts(self):
        ui = self.physics_manager.user_info

        # range
        if ui.energy_range_min is not None and ui.energy_range_max is not None:
            self.physics_manager.warn_user("SetEnergyRange only works in MT mode")
            pct = g4.G4ProductionCutsTable.GetProductionCutsTable()
            pct.SetEnergyRange(ui.energy_range_min, ui.energy_range_max)

        # Set global production cuts
        # If value is set for 'all', this overrides individual values
        if ui.global_production_cuts.all is not None:
            # calls SetCutValue for all relevant particles,
            # i.e. proton, gamma, e+, e-
            for pname in cut_particle_names.values():
                self.g4_physics_list.SetCutValue(ui.global_production_cuts.all, pname)

        else:
            for pname, value in ui.global_production_cuts.items():
                # ignore 'all', as that's already treated above
                if pname == "all":
                    continue
                if value is not None and value not in ("default", "Default"):
                    self.g4_physics_list.SetCutValue(
                        value, translate_particle_name_gate_to_geant4(pname)
                    )

    def initialize_g4_em_parameters(self):
        self.g4_em_parameters = g4.G4EmParameters.Instance()

        self.g4_em_parameters.SetApplyCuts(self.physics_manager.apply_cuts)

        if self.physics_manager.em_parameters.fluo is not None:
            self.g4_em_parameters.SetFluo(self.physics_manager.em_parameters.fluo)
        if self.physics_manager.em_parameters.auger is not None:
            self.g4_em_parameters.SetAuger(self.physics_manager.em_parameters.auger)
        if self.physics_manager.em_parameters.auger_cascade is not None:
            self.g4_em_parameters.SetAugerCascade(
                self.physics_manager.em_parameters.auger_cascade
            )
        if self.physics_manager.em_parameters.pixe is not None:
            self.g4_em_parameters.SetPixe(self.physics_manager.em_parameters.pixe)
        if self.physics_manager.em_parameters.deexcitation_ignore_cut is not None:
            self.g4_em_parameters.SetDeexcitationIgnoreCut(
                self.physics_manager.em_parameters.deexcitation_ignore_cut
            )

        # set the deex switches only if the user has touched them.
        # Let G4 set its defaults otherwise (i.e. of all are None)
        if any(
            [v is not None for v in self.physics_manager.em_switches_world.values()]
        ):
            # check that all switches were set in case at least one has been set
            # either all must be set or none
            if any(
                [v is None for v in self.physics_manager.em_switches_world.values()]
            ):
                fatal(
                    f"Some EM switches for the world region were not set. You must either set all switches or none. The following switches exist: {self.physics_manager.em_switches_world.keys()}"
                )
            self.g4_em_parameters.SetDeexActiveRegion(
                "world",
                self.physics_manager.em_switches_world.deex,
                self.physics_manager.em_switches_world.auger,
                self.physics_manager.em_switches_world.pixe,
            )
        for region in self.physics_manager.regions.values():
            region.initialize_em_switches()

    def initialize_physics_biasing(self):
        # get a dictionary {particle:[processes]}
        particles_processes = self.physics_manager.get_biasing_particles_and_processes()

        # check if there are any processes requested for any particle
        if any([len(v) > 0 for v in particles_processes.values()]):
            g4_biasing_physics = g4.G4GenericBiasingPhysics()
            for particle, processes in particles_processes.items():
                if len(processes) > 0:
                    g4_biasing_physics.PhysicsBias(particle, processes)
            self.g4_physics_list.RegisterPhysics(g4_biasing_physics)

    # This function deals with calling the parse function
    # and setting the returned MaterialPropertyTable to G4Material object
    def initialize_optical_material_properties(self):
        # Load optical material properties if special physics constructor "G4OpticalPhysics"
        # is set to True in PhysicsManager's user info
        if (
            self.simulation_engine.simulation.physics_manager.special_physics_constructors.G4OpticalPhysics
            is True
        ):
            # retrieve path to file from physics manager
            for (
                vol
            ) in self.simulation_engine.simulation.volume_manager.volumes.values():
                material_name = vol.g4_material.GetName()
                material_properties = load_optical_properties_from_xml(
                    self.physics_manager.optical_properties_file, material_name
                )
                if material_properties is not None:
                    self.g4_optical_material_tables[str(material_name)] = (
                        create_g4_optical_properties_table(material_properties)
                    )
                    vol.g4_material.SetMaterialPropertiesTable(
                        self.g4_optical_material_tables[str(material_name)]
                    )
                else:
                    self.simulation_engine.simulation.warn_user(
                        f"Could not load the optical material properties for material {material_name} "
                        f"found in volume {vol.name} from file {self.physics_manager.optical_properties_file}."
                    )

    @requires_fatal("physics_manager")
    def initialize_optical_surfaces(self):
        """Calls initialize() method of each OpticalSurface instance."""

        # Call the initialize() method in OpticalSurface class to
        # create the related G4 instances.
        for optical_surface in self.physics_manager.optical_surfaces.values():
            optical_surface.initialize()

    @requires_fatal("physics_manager")
    def initialize_user_limits_physics(self):
        need_step_limiter = False
        need_user_special_cut = False
        for r in self.physics_manager.regions.values():
            if r.need_step_limiter() is True:
                need_step_limiter = True
            if r.need_user_special_cut() is True:
                need_user_special_cut = True

        if need_step_limiter or need_user_special_cut:
            user_limits_physics = UserLimitsPhysics()
            user_limits_physics.physics_engine = self
            self.g4_physics_list.RegisterPhysics(user_limits_physics)
            self.gate_physics_constructors.append(user_limits_physics)

    @requires_fatal("physics_manager")
    def initialize_ionisation_options(self):
        for material_name, val in self.physics_manager.mean_energy_per_ion_pair.items():
            mat = (
                self.simulation_engine.simulation.volume_manager.find_or_build_material(
                    material_name
                )
            )
            ionisation = mat.GetIonisation()
            ionisation.SetMeanEnergyPerIonPair(val)

    @requires_fatal("physics_manager")
    def initialize_users_ionisation_potentials(self):
        for (
            material_name,
            val,
        ) in self.physics_manager.material_ionisation_potential.items():
            mat = (
                self.simulation_engine.simulation.volume_manager.find_or_build_material(
                    material_name
                )
            )
            ionisation = mat.GetIonisation()
            ionisation.SetMeanExcitationEnergy(val)


class ActionEngine(g4.G4VUserActionInitialization, EngineBase):
    """
    Main object to manage all actions during a simulation.
    """

    def __init__(self, *args):
        g4.G4VUserActionInitialization.__init__(self)
        EngineBase.__init__(self, *args)

        # *** G4 references ***
        # List of G4 source managers (one per thread)
        self.g4_PrimaryGenerator = []

        # The main G4 source manager
        self.g4_main_PrimaryGenerator = None

        # Lists of elements to prevent destruction
        self.g4_RunAction = []
        self.g4_EventAction = []
        self.g4_TrackingAction = []

    def close(self):
        if self.verbose_close:
            warning("Closing ActionEngine")
        self.release_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_PrimaryGenerator = []
        self.g4_main_PrimaryGenerator = None
        self.g4_RunAction = []
        self.g4_EventAction = []
        self.g4_TrackingAction = []

    def register_all_actions(self, actor):
        self.register_run_actions(actor)
        self.register_event_actions(actor)
        self.register_tracking_actions(actor)

    def register_run_actions(self, actor):
        for ra in self.g4_RunAction:
            ra.RegisterActor(actor)

    def register_event_actions(self, actor):
        for ea in self.g4_EventAction:
            ea.RegisterActor(actor)

    def register_tracking_actions(self, actor):
        for ta in self.g4_TrackingAction:
            ta.RegisterActor(actor)

    def BuildForMaster(self):
        # This function is called only in MT mode, for the master thread
        if not self.g4_main_PrimaryGenerator:
            self.g4_main_PrimaryGenerator = (
                self.simulation_engine.source_engine.create_master_source_manager()
            )

    def Build(self):
        # In MT mode this Build function is invoked
        # for each worker thread, so all user action classes
        # are defined thread-locally.

        # If MT is not enabled, need to create the main source
        if not self.g4_main_PrimaryGenerator:
            p = self.simulation_engine.source_engine.create_master_source_manager()
            self.g4_main_PrimaryGenerator = p
        else:
            # else create a source for each thread
            p = self.simulation_engine.source_engine.create_g4_thread_source_manager()

        self.SetUserAction(p)
        self.g4_PrimaryGenerator.append(p)

        # set the actions for Run
        ra = g4.GateRunAction(p)
        self.SetUserAction(ra)
        self.g4_RunAction.append(ra)

        # set the actions for Event
        ea = g4.GateEventAction()
        self.SetUserAction(ea)
        self.g4_EventAction.append(ea)

        # set the actions for Track
        ta = g4.GateTrackingAction()
        ta.fUserEventInformationFlag = (
            self.simulation_engine.user_event_information_flag
        )
        self.SetUserAction(ta)
        self.g4_TrackingAction.append(ta)


def register_sensitive_detector_to_children(actor, lv):
    logger.debug(
        f'Actor: "{actor.user_info.name}" '
        f'(attached to "{actor.attached_to}") '
        f'set to volume "{lv.GetName()}"'
    )
    actor.RegisterSD(lv)
    for i in range(lv.GetNoDaughters()):
        child = lv.GetDaughter(i).GetLogicalVolume()
        register_sensitive_detector_to_children(actor, child)


class ActorEngine(EngineBase):
    """
    This object manages all actors G4 objects at runtime
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_to_actors()

    @property
    def actor_manager(self):
        return self.simulation_engine.simulation.actor_manager

    def close(self):
        if self.verbose_close:
            warning("Closing ActorEngine")
        for actor in self.actor_manager.actors.values():
            actor.close()
        super().close()

    def initialize(self):
        for actor in self.actor_manager.sorted_actors:
            logger.debug(f"Actor: initialize [{actor.type_name}] {actor.name}")
            self.simulation_engine.action_engine.register_all_actions(actor)
            actor.initialize()
            # warning: the step actions will be registered by register_sensitive_detectors
            # called by ConstructSDandField

    def register_to_actors(self):
        for actor in self.actor_manager.actors.values():
            actor.actor_engine = self

    def register_sensitive_detectors(self, world_name):
        for actor in self.actor_manager.sorted_actors:
            if actor.IsSensitiveDetector() is True:
                # Step: only enabled if attachTo a given volume.
                # Propagated to all child and sub-child
                # tree = volume_manager.volumes_tree
                if isinstance(actor.attached_to, str):
                    # make a list with one single element
                    mothers = [actor.attached_to]
                else:
                    mothers = actor.attached_to
                # add SD for all mothers
                for volume_name in mothers:
                    volume = (
                        self.simulation_engine.simulation.volume_manager.get_volume(
                            volume_name
                        )
                    )
                    if volume.world_volume.name == world_name:
                        register_sensitive_detector_to_children(
                            actor, volume.g4_logical_volume
                        )

            # this is specific for BiasingOperator/Actor
            # this is needed for MultiThread run
            if hasattr(actor, "ConfigureForWorker"):
                actor.InitializeUserInfo(actor.user_info)
                actor.ConfigureForWorker()

    def start_simulation(self):
        # consider the priority value of the actors
        for actor in self.actor_manager.sorted_actors:
            actor.StartSimulationAction()

    def stop_simulation(self):
        # consider the priority value of the actors
        for actor in self.actor_manager.sorted_actors:
            actor.EndSimulationAction()


class FilterEngine(EngineBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def filter_manager(self):
        return self.simulation_engine.simulation.filter_manager

    def initialize(self):
        for f in self.filter_manager.filters.values():
            f.initialize()

    def close(self):
        for f in self.filter_manager.filters.values():
            f.close()
        super().close()


class ParallelWorldEngine(g4.G4VUserParallelWorld, EngineBase):
    """FIXME: Doc ParallelWorldEngine"""

    def __init__(self, parallel_world_name, *args):
        g4.G4VUserParallelWorld.__init__(self, parallel_world_name)
        EngineBase.__init__(self, *args)

        # keep input data
        self.parallel_world_name = parallel_world_name
        # the parallel world volume needs the engine to construct itself
        self.parallel_world_volume.parallel_world_engine = self

    def close(self):
        self.parallel_world_volume.parallel_world_engine = None
        super().close()

    @property
    def parallel_world_volume(self):
        return (
            self.simulation_engine.volume_engine.volume_manager.parallel_world_volumes[
                self.parallel_world_name
            ]
        )

    def Construct(self):
        """
        G4 overloaded.
        Override the Construct method from G4VUserParallelWorld
        """

        # Construct all volumes within this world along the tree hierarchy
        # The world volume of this world is the first item
        for volume in PreOrderIter(self.parallel_world_volume):
            volume.construct()

    def ConstructSD(self):
        # FIXME
        self.simulation_engine.actor_engine.register_sensitive_detectors(
            self.parallel_world_name,
        )


class VolumeEngine(g4.G4VUserDetectorConstruction, EngineBase):
    """
    Engine that will create all G4 elements for the hierarchy of volumes.
    Correspond to the G4VUserDetectorConstruction (inherit)
    Also manage the list of parallel worlds.
    """

    def __init__(self, simulation_engine):
        g4.G4VUserDetectorConstruction.__init__(self)
        EngineBase.__init__(self, simulation_engine)

        self._is_constructed = False

        self.volume_manager = self.simulation_engine.simulation.volume_manager

        # parallel world engines will be created by the simulation engine
        self.parallel_world_engines = {}

        # Make sure all volumes have a reference to this engine
        self.register_to_volumes()

    def create_parallel_world_engines(self):
        for parallel_world_name in self.volume_manager.parallel_world_names:
            self.parallel_world_engines[parallel_world_name] = ParallelWorldEngine(
                parallel_world_name, self.simulation_engine
            )
            self.RegisterParallelWorld(self.parallel_world_engines[parallel_world_name])

    def register_to_volumes(self):
        for v in self.volume_manager.volumes.values():
            v.volume_engine = self

    def close(self):
        for vol in self.volume_manager.volumes.values():
            vol.close()
        for pwv in self.volume_manager.parallel_world_volumes.values():
            pwv.close()
        for pwe in self.parallel_world_engines.values():
            pwe.close()
        self.parallel_world_engines = {}
        super().close()

    def initialize(self):
        # build the materials
        self.simulation_engine.simulation.volume_manager.material_database.initialize()
        # initialize actors which handle dynamic volume parametrization, e.g. MotionActors
        self.initialize_dynamic_parametrisations()

    def initialize_dynamic_parametrisations(self):
        dynamic_volumes = self.volume_manager.dynamic_volumes
        for vol in self.volume_manager.dynamic_volumes:
            vol.check_if_dynamic_params_match_run_timing_intervals()
        if len(dynamic_volumes) > 0:
            dynamic_geometry_actor = self.simulation_engine.simulation.add_actor(
                "DynamicGeometryActor", "dynamic_geometry_actor"
            )
            dynamic_geometry_actor.priority = 0
            for vol in self.volume_manager.dynamic_volumes:
                dynamic_geometry_actor.geometry_changers.extend(vol.create_changers())

    def Construct(self):
        """
        G4 overloaded.
        Override the Construct method from G4VUserDetectorConstruction
        """

        # # build the materials
        # # FIXME: should go into initialize method
        # self.simulation_engine.simulation.volume_manager.material_database.initialize()

        # Construct all volumes within the mass world along the tree hierarchy
        # The world volume is the first item

        self.volume_manager.update_volume_tree()
        for volume in PreOrderIter(self.volume_manager.world_volume):
            volume.construct()

        # return the (main) world physical volume
        self._is_constructed = True
        return self.volume_manager.world_volume.g4_physical_volume

    def check_overlaps(self, verbose):
        for volume in self.volume_manager.volumes.values():
            if volume not in self.volume_manager.all_world_volumes:
                for pw in volume.g4_physical_volumes:
                    try:
                        b = pw.CheckOverlaps(1000, 0, verbose, 1)
                        if b:
                            fatal(
                                f'Some volumes overlap with the volume "{volume.name}". \n'
                                f"Use Geant4's verbose output to know which ones. \n"
                                f"Aborting."
                            )
                    # FIXME: What causes the exceptions?
                    except:
                        warning(f"Could not check overlap for volume {volume.name}.")

    def ConstructSDandField(self):
        """
        G4 overloaded
        """
        # This function is called in MT mode
        self.simulation_engine.actor_engine.register_sensitive_detectors(
            self.volume_manager.world_volume.name,
        )

    def get_volume(self, name):
        return self.volume_manager.get_volume(name)

    def get_database_material_names(self, db=None):
        return self.simulation_engine.simulation.volume_manager.material_database.get_database_material_names(
            db
        )

    def dump_build_materials(self, level=0):
        table = g4.G4Material.GetMaterialTable
        if level == 0:
            names = [m.GetName() for m in table]
            return names
        return table


class VisualisationEngine(EngineBase):
    """
    Main class to manage visualisation
    """

    def __init__(self, simulation_engine):
        EngineBase.__init__(self, simulation_engine)
        self.g4_vis_executive = None
        self.current_visu_filename = None
        self._is_closed = None
        self.simulation = simulation_engine.simulation

    def close(self):
        if self.simulation_engine.verbose_close:
            warning(f"Closing VisualisationEngine is_closed = {self._is_closed}")
        # self.release_g4_references()
        self._is_closed = True

    def release_g4_references(self):
        self.g4_vis_executive = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def initialize_visualisation(self):
        # check if filename is set when needed
        if (
            "only" in self.simulation.visu_type
            and self.simulation.visu_filename is None
        ):
            fatal(
                f'You must define a visu_filename with "{self.simulation.visu_type}" is set'
            )

        # set the current filename (maybe changed is no visu_filename)
        self.current_visu_filename = self.simulation.visu_filename

        # gdml
        if self.simulation.visu_type in ("gdml", "gdml_file_only"):
            self.initialize_visualisation_gdml()

        # vrml
        if self.simulation.visu_type in ("vrml", "vrml_file_only"):
            self.initialize_visualisation_vrml()

        # G4 stuff
        self.g4_vis_executive = g4.G4VisExecutive("all")
        self.g4_vis_executive.Initialize()

    def initialize_visualisation_gdml(self):
        # Check when GDML is activated, if G4 was compiled with GDML
        if not g4.GateInfo.get_G4GDML():
            self.simulation.warn_user(
                "Visualization with GDML not available in Geant4. Check G4 compilation."
            )
        if self.current_visu_filename is None:
            self.current_visu_filename = f"gate_visu_{os.getpid()}.gdml"

    def initialize_visualisation_vrml(self):
        if self.simulation.visu_filename is not None:
            os.environ["G4VRMLFILE_FILE_NAME"] = self.simulation.visu_filename
        else:
            self.current_visu_filename = f"gate_visu_{os.getpid()}.wrl"
            os.environ["G4VRMLFILE_FILE_NAME"] = self.current_visu_filename

    def start_visualisation(self):
        if not self.simulation.visu:
            return

        # VRML ?
        if self.simulation.visu_type == "vrml":
            start_vrml_visu(self.current_visu_filename)

        # GDML ?
        if self.simulation.visu_type == "gdml":
            start_gdml_visu(self.current_visu_filename)

        # remove the temporary file
        if self.simulation.visu_filename is None:
            try:
                os.remove(self.current_visu_filename)
            except:
                pass


class SimulationOutput:
    """
    FIXME
    """

    def __init__(self):
        self.simulation = None
        self.actors = {}
        self.sources = {}
        self.sources_by_thread = {}
        self.pid = os.getpid()
        self.ppid = os.getppid()
        self.current_random_seed = None
        self.user_hook_log = []
        self.warnings = None

    def store_actors(self, simulation_engine):
        self.actors = simulation_engine.simulation.actor_manager.actors
        for actor in self.actors.values():
            actor.close()

    def store_hook_log(self, simulation_engine):
        self.user_hook_log = simulation_engine.user_hook_log

    def store_sources(self, simulation_engine):
        self.sources = {}
        if simulation_engine.simulation.multithreaded is True:
            th = {}
            self.sources_by_thread = [
                {} for _ in range(simulation_engine.simulation.number_of_threads + 1)
            ]
            for source in simulation_engine.source_engine.sources:
                n = source.name
                if n in th:
                    th[n] += 1
                else:
                    th[n] = 0
                self.sources_by_thread[th[n]][n] = source
            self.sources = self.sources_by_thread[0]
        else:
            s = {}
            for source in simulation_engine.source_engine.sources:
                s[source.name] = source
            self.sources = s

    def get_actor(self, name):
        if name not in self.actors:
            fatal(
                f'The actor "{name}" does not exist. '
                f"These are the actors known to this simulation: {list(self.actors.keys())}"
            )
        return self.actors[name]

    def get_source(self, name):
        if name not in self.sources:
            fatal(
                f'The source "{name}" does not exist. Here is the list of sources: {list(self.sources.keys())}'
            )
        return self.sources[name]

    def get_source_mt(self, name, thread):
        if (
            self.simulation.number_of_threads <= 1
            and not self.simulation.force_multithread_mode
        ):
            fatal("Cannot use get_source_mt in monothread mode")
        if thread >= len(self.sources_by_thread):
            fatal(
                f"Cannot get source {name} with thread {thread}, while "
                f"there are only {len(self.sources_by_thread)} threads"
            )
        if name not in self.sources_by_thread[thread]:
            s = self.sources_by_thread[thread].keys
            fatal(
                f'The source "{name}" does not exist. Here is the list of sources: {s}'
            )
        return self.sources_by_thread[thread][name]


class SimulationEngine(GateSingletonFatal):
    """
    Main class to execute a Simulation. Can only be created once per process because Geant4 dictates so.
    """

    def __init__(self, simulation, new_process=False):
        self.simulation = simulation
        self.verbose_getstate = simulation.verbose_getstate
        self.verbose_close = simulation.verbose_close

        # create engines passing the simulation engine (self) as argument
        self.volume_engine = VolumeEngine(self)
        self.volume_engine.create_parallel_world_engines()
        self.physics_engine = PhysicsEngine(self)
        self.source_engine = SourceEngine(self)
        self.action_engine = ActionEngine(self)
        self.actor_engine = ActorEngine(self)
        self.filter_engine = FilterEngine(self)
        self.visu_engine = VisualisationEngine(self)

        # current state of the engine
        self.run_timing_intervals = None
        self.is_initialized = False

        # do we create a subprocess or not ?
        # this is only for info.
        # process handling is done in Simulation class, not in SimulationEngine!
        self.new_process = new_process

        # LATER : option to wait the end of completion or not

        # UI
        self.g4_ui_session = None
        self.g4_ui = None

        # random engine
        self.g4_HepRandomEngine = None
        self.current_random_seed = None

        # Main Run Manager
        self.g4_RunManager = None
        self.g4_StateManager = g4.G4StateManager.GetStateManager()

        # life cycle management
        self.run_manager_finalizer = None
        self._is_closed = False

        # exception handler
        self.g4_exception_handler = None

        # user fct to call after initialization
        self.user_hook_after_init = simulation.user_hook_after_init
        self.user_hook_after_init_arg = simulation.user_hook_after_init_arg
        self.user_hook_after_run = simulation.user_hook_after_run
        # a list to store short log messages
        # produced by hook function such as user_hook_after_init
        self.user_hook_log = []  # FIXME: turn this into dictionary

    def close_engines(self):
        if self.volume_engine:
            self.volume_engine.close()
        if self.physics_engine:
            self.physics_engine.close()
        if self.source_engine:
            self.source_engine.close()
        if self.action_engine:
            self.action_engine.close()
        if self.actor_engine:
            self.actor_engine.close()
        if self.visu_engine:
            self.visu_engine.close()

    def release_engines(self):
        self.volume_engine = None
        self.physics_engine = None
        self.source_engine = None
        self.action_engine = None
        self.actor_engine = None
        self.visu_engine = None

    def release_g4_references(self):
        self.g4_ui_session = None
        self.g4_ui = None
        self.g4_HepRandomEngine = None
        self.g4_StateManager = None
        self.g4_exception_handler = None

    def notify_managers(self):
        self.simulation.physics_manager._simulation_engine_closing()
        self.simulation.volume_manager._simulation_engine_closing()

    def close(self):
        if self.verbose_close:
            warning(f"Closing SimulationEngine is_closed = {self._is_closed}")
        if self._is_closed is False:
            self.close_engines()
            self.release_engines()
            self.release_g4_references()
            self.notify_managers()
            if self.g4_RunManager:
                self.g4_RunManager.SetVerboseLevel(0)
            self._is_closed = True
        self.g4_RunManager = None

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def __getstate__(self):
        raise GateImplementationError(
            "The __getstate__() method of the SimulationEngine class got called. "
            "That should not happen because it should be closed before anything is pickled "
            "at the end of a subprocess. Check warning messages for clues!"
        )
        # if self.simulation.verbose_getstate:
        #     warning("Getstate SimulationEngine")
        # self.g4_StateManager = None
        # # if self.user_fct_after_init is not None:
        # #    gate.warning(f'Warning')
        # return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d
        # recreate the StateManager when unpickling
        # because it was set to None during pickling
        self.g4_StateManager = g4.G4StateManager.GetStateManager()

    def run_engine(self):
        """
        When the simulation is about to init, if the Simulation object is in a separate process
        (with 'spawn'), it has been pickled (copied), and the G4 phys list class does not exist
        any more. So we need to recreate them with 'create_physics_list_classes'
        Also, the StateManager must be recreated.

        NK: Yes, but not this way. Each class should take care of recreating attributes
        which where set to None during pickling by implementing a __setstate__ method.
        Implementing the resetting somewhere else (maybe in multiple places...) in the code will
        make it very difficult to maintain.

        -> removed the lines and implemented __setstate__ methods for the classes in question
        """

        # initialise the logger and its output
        original_stdout = self.simulation.initialize_logger()

        # prepare the output
        output = SimulationOutput()

        # if the simulation is run in a subprocess,
        # we want to capture only the warnings from this point on
        # because everything else has already been executed in the main process
        # and potential warnings have already been registered.
        # if self.new_process is True:
        #    self.simulation.reset_warnings()

        # initialization
        self.initialize()

        # things to do after init and before run
        self.apply_all_g4_commands_after_init()

        if self.user_hook_after_init:
            logger.info("Simulation: initialize user fct")
            if self.user_hook_after_init_arg is not None:
                self.user_hook_after_init(self, self.user_hook_after_init_arg)
            else:
                self.user_hook_after_init(self)

        # if init only, we stop
        if self.simulation.init_only:
            output.store_actors(self)
            output.store_sources(self)
            output.store_hook_log(self)
            output.current_random_seed = self.current_random_seed
            output.expected_number_of_events = (
                self.source_engine.expected_number_of_events
            )
            output.warnings = self.simulation.warnings
            output.log_output = self.simulation.log_output.getvalue()
            sys.stdout = original_stdout
            return output

        # go
        self.start_and_stop()

        # start visualisation if vrml or gdml
        self.visu_engine.start_visualisation()
        if self.user_hook_after_run:
            logger.info("Simulation: User hook after run")
            self.user_hook_after_run(self)

        # prepare the output
        output.store_actors(self)
        output.store_sources(self)
        output.store_hook_log(self)
        output.current_random_seed = self.current_random_seed
        output.expected_number_of_events = self.source_engine.expected_number_of_events
        output.warnings = self.simulation.warnings

        # save the output and restore the sys.stdout
        output.log_output = self.simulation.log_output.getvalue()
        sys.stdout = original_stdout

        return output

    def start_and_stop(self):
        """
        Start the simulation. The runs are managed in the SourceManager.
        """

        s = ""
        if self.new_process:
            s = "(in a new process) "
        s2 = ""
        if self.simulation.progress_bar:
            n = self.source_engine.expected_number_of_events
            if self.source_engine.can_predict_expected_number_of_event():
                s2 = f"(around {n} events expected)"
            else:
                s2 = f"(cannot predict the number of events, max is {n}, e.g. acceptance_angle is enabled)"

        logger.info("-" * 80 + f"\nSimulation: START {s}{s2}")

        # actor: start simulation (only the master thread)
        self.actor_engine.start_simulation()

        # go !
        start = time.time()
        self.source_engine.start()
        end = time.time()

        # actor: stop simulation (only the master thread)
        self.actor_engine.stop_simulation()

        # this is the end
        logger.info(
            f"Simulation: STOP. Run: {len(self.run_timing_intervals)}. "
            # f'Events: {self.source_manager.total_events_count}. '
            f"Time: {end - start:0.1f} seconds.\n"
            + "-" * 80
        )

    def initialize_random_engine(self):
        engine_name = self.simulation.random_engine
        self.g4_HepRandomEngine = None
        if engine_name == "MixMaxRng":
            self.g4_HepRandomEngine = g4.MixMaxRng()
        if engine_name == "MersenneTwister":
            self.g4_HepRandomEngine = g4.MTwistEngine()
        if not self.g4_HepRandomEngine:
            s = f"Cannot find the random engine {engine_name}\n"
            s += "Use: MersenneTwister or MixMaxRng"
            fatal(s)

        # set the random engine
        g4.G4Random.setTheEngine(self.g4_HepRandomEngine)
        if self.simulation.random_seed == "auto":
            self.current_random_seed = random.randrange(sys.maxsize)
        else:
            self.current_random_seed = self.simulation.random_seed

        # if windows, the long are 4 bytes instead of 8 bytes for python and unix system
        if os.name == "nt":
            self.current_random_seed = int(
                self.current_random_seed % ((pow(2, 32) - 1) / 2)
            )

        # set the seed
        g4.G4Random.setTheSeed(self.current_random_seed, 0)

    def initialize_g4_verbose(self):
        if self.simulation.g4_verbose:
            # Geant4 output with color
            ui = UIsessionVerbose()
            # set verbose tracking according to user info:
        else:
            # no Geant4 output
            ui = UIsessionSilent()
        if self.simulation.g4_verbose_level_tracking >= 0:
            self.simulation.g4_commands_after_init.append(
                f"/tracking/verbose {self.simulation.g4_verbose_level_tracking}"
            )
        # it is also possible to set ui=None for 'default' output
        # we must keep a ref to ui_session
        self.g4_ui_session = ui
        # we must keep a ref to ui_manager
        self.g4_ui = g4.G4UImanager.GetUIpointer()
        if self.g4_ui is None:
            fatal("Unable to obtain a UIpointer")
        self.g4_ui.SetCoutDestination(ui)

    def initialize(self):
        """
        Build the main geant4 objects and initialize them.
        """

        # g4 verbose
        self.initialize_g4_verbose()

        # visualisation ?
        # self.pre_init_visu()

        # init random engine (before the MTRunManager creation)
        self.initialize_random_engine()

        # Some sources (e.g. PHID) need to perform computation once everything is defined in user_info but *before* the
        # initialization of the G4 engine starts. This can be done via this function.
        self.simulation.initialize_source_before_g4_engine()

        # create the run manager (assigned to self.g4_RunManager)
        if self.g4_RunManager:
            fatal("A G4RunManager as already been created.")
        self.g4_RunManager = self.create_run_manager()
        # this creates a finalizer for the run manager which assures that
        # the close() method is called before the run manager is garbage collected,
        # i.e. G4RunManager destructor is called
        self.run_manager_finalizer = weakref.finalize(self.g4_RunManager, self.close)

        # create the handler for the exception
        self.g4_exception_handler = ExceptionHandler()

        # check run timing
        self.run_timing_intervals = self.simulation.run_timing_intervals.copy()
        # FIXME: put this assertion in a setter hook
        assert_run_timing(self.run_timing_intervals)

        # check if some actors need UserEventInformation
        # FIXME: should go to ActorEngine
        self.initialize_user_event_information_flag()

        # Geometry initialization
        logger.info("Simulation: initialize Geometry")
        self.volume_engine.initialize()

        # Physics initialization
        logger.info("Simulation: initialize Physics")
        self.physics_engine.initialize_before_runmanager()

        # Apply G4 commands *before* init (after phys init)
        self.apply_all_g4_commands_before_init()

        # sources
        logger.info("Simulation: initialize Source")
        self.source_engine.initialize(
            self.simulation.run_timing_intervals, self.simulation.progress_bar
        )

        # action

        # Visu
        if self.simulation.visu:
            logger.info("Simulation: initialize Visualization")
            self.visu_engine.initialize_visualisation()

        # set pointers to python classes
        self.g4_RunManager.SetUserInitialization(self.volume_engine)
        self.g4_RunManager.SetUserInitialization(self.physics_engine.g4_physics_list)
        self.g4_RunManager.SetUserInitialization(
            self.action_engine
        )  # G4 internally calls action_engine.Build()

        # Important: The volumes are constructed
        # when the G4RunManager calls the Construct method of the VolumeEngine,
        # which happens in the InitializeGeometry() method of the
        # G4RunManager within Initialize() (Geant4 code)

        # Note: In serial mode, SetUserInitialization() would only be needed
        # for geometry and physics, but in MT mode the fake run for worker
        # initialization needs a particle source.
        logger.info("Simulation: initialize G4RunManager")
        if self.simulation.multithreaded is True:
            self.g4_RunManager.InitializeWithoutFakeRun()
        else:
            self.g4_RunManager.Initialize()

        logger.info(
            "Simulation: initialize PhysicsEngine after RunManager initialization"
        )
        self.physics_engine.initialize_after_runmanager()
        self.g4_RunManager.PhysicsHasBeenModified()

        # G4's MT RunManager needs an empty run to initialise workers
        if self.simulation.multithreaded is True:
            self.g4_RunManager.FakeBeamOn()

        # Actions initialisation
        # This must come after the G4RunManager initialisation
        # because the RM initialisation calls ActionEngine.Build()
        # which is required for initialize()
        # Actors initialization (before the RunManager Initialize)
        # self.actor_engine.create_actors()  # calls the actors' constructors
        logger.info("Simulation: initialize actors")
        self.source_engine.initialize_actors()
        self.actor_engine.initialize()
        self.filter_engine.initialize()

        self.is_initialized = True

        # Check overlaps
        if self.simulation.check_volumes_overlap:
            logger.info("Simulation: check volumes overlap")
            self.check_volumes_overlap(verbose=False)
        else:
            logger.info("Simulation: (no volumes overlap checking)")

        # Register sensitive detector.
        # If G4 was compiled with MT (regardless if it is used or not),
        # ConstructSDandField (in VolumeManager) will be automatically called
        if not g4.GateInfo.get_G4MULTITHREADED():
            fatal("DEBUG Register sensitive detector in no MT mode")
            # todo : self.actor_engine.register_sensitive_detectors()

    def create_run_manager(self):
        """Get the correct RunManager according to the requested threads
        and make some basic settings.

        """
        if self.simulation.multithreaded is True:
            # GetOptions() returns a set which should contain 'MT'
            # if Geant4 was compiled with G4MULTITHREADED
            if "MT" not in g4.G4RunManagerFactory.GetOptions():
                fatal(
                    "Geant4 does not support multithreading. Probably it was compiled without G4MULTITHREADED flag."
                )

            logger.info(
                f"Simulation: create MTRunManager with {self.simulation.number_of_threads} threads"
            )
            g4_RunManager = g4.WrappedG4MTRunManager()
            g4_RunManager.SetNumberOfThreads(self.simulation.number_of_threads)
        else:
            logger.info("Simulation: create RunManager (single thread)")
            g4_RunManager = g4.WrappedG4RunManager()

        if g4_RunManager is None:
            fatal("Unable to create RunManager")

        g4_RunManager.SetVerboseLevel(self.simulation.g4_verbose_level)
        return g4_RunManager

    def apply_all_g4_commands_after_init(self):
        for command in self.simulation.g4_commands_after_init:
            self.add_g4_command_after_init(command)

    def apply_all_g4_commands_before_init(self):
        for command in self.simulation.g4_commands_before_init:
            self.add_g4_command_after_init(command)

    def add_g4_command_after_init(self, command):
        if self.g4_ui is None:
            self.g4_ui = g4.G4UImanager.GetUIpointer()
        logger.info(f"Simulation: apply G4 command '{command}'")
        code = self.g4_ui.ApplyCommand(command)
        if code == 0:
            return
        err_codes = {
            0: "fCommandSucceeded",
            100: "fCommandNotFound",
            200: "fIllegalApplicationState",
            300: "fParameterOutOfRange",
            400: "fParameterUnreadable",
            500: "fParameterOutOfCandidates",
            600: "fAliasNotFound",
        }
        closest_err_code = max([x for x in err_codes if x <= code])
        closest_err_msg = err_codes[closest_err_code]
        fatal(f'Error in apply_g4_command "{command}": {code} {closest_err_msg}')

    def check_volumes_overlap(self, verbose=True):
        # we need to 'cheat' the verbosity before doing the check
        b = self.simulation.g4_verbose
        self.simulation.g4_verbose = True
        self.initialize_g4_verbose()

        # check
        self.volume_engine.check_overlaps(verbose)

        # put back verbosity
        self.simulation.g4_verbose = b
        self.initialize_g4_verbose()

    @property
    @requires_fatal("g4_StateManager")
    def g4_state(self):
        return self.g4_StateManager.GetCurrentState()

    @g4_state.setter
    @requires_fatal("g4_StateManager")
    def g4_state(self, g4_application_state):
        self.g4_StateManager.SetNewState(g4_application_state)

    def initialize_user_event_information_flag(self):
        self.user_event_information_flag = False
        for ac in self.simulation.actor_manager.actors.values():
            if "attributes" in ac.user_info:
                if "ParentParticleName" in ac.attributes:
                    self.user_event_information_flag = True


def start_gdml_visu(filename):
    try:
        import pyg4ometry
    except Exception as exception:
        warning(exception)
        warning(
            "The module pyg4ometry is maybe not installed or is not working. Try: \n"
            "pip install pyg4ometry"
        )
        return
    r = pyg4ometry.gdml.Reader(filename)
    l = r.getRegistry().getWorldVolume()
    v = pyg4ometry.visualisation.VtkViewerColouredMaterial()
    v.addLogicalVolume(l)
    v.view()


def start_vrml_visu(filename):
    try:
        import pyvista
    except Exception as exception:
        warning(exception)
        warning(
            "The module pyvista is maybe not installed or is not working to be able to visualize vrml files. Try:\n"
            "pip install pyvista"
        )
        return
    pl = pyvista.Plotter()
    pl.import_vrml(filename)
    pl.set_background("black")
    pl.add_axes(line_width=5, color="white")
    pl.show()



import numpy as np
import scipy as sc
from numpy.random import MT19937
from numpy.random import RandomState, SeedSequence
import random
from box import Box
import textwrap
import inspect
import importlib.resources as resources
import sys
from pathlib import Path
import string
import os
import re
import json
import importlib
import importlib.util
from importlib.metadata import version
import shutil

import opengate_core as g4
from opengate import get_site_packages_dir
from .exception import fatal, warning


class LazyModuleLoader:
    """
    Lazy loading allows you to delay the loading of a module until it's actually needed.
    This can be useful if a module is expensive to load or if it may not be used in
    every execution of the program.
    We use it for some modules that was found to delay the startup time, in particular:
    - radioactivedecay and pandas in phidsources
    - torch and gaga in gansources (only required for some features)
    """

    def __init__(self, module_name):
        self.module_name = module_name
        self.module = None

    def __getattr__(self, name):
        if self.module is None:
            # Check module existence and import it
            try:
                # print(f"LazyModuleLoader is importing module {self.module_name} ...")
                self.module = importlib.import_module(self.module_name)
                # print("... done")
            except ModuleNotFoundError:
                fatal(
                    f"The module '{self.module_name}' is not installed. "
                    f"Please install it before proceeding."
                )
        return getattr(self.module, name)


git = LazyModuleLoader("git")


def assert_equal_dic(d1, d2, name=""):
    for k in d1:
        if k not in d2:
            fatal(f"ERROR missing key {k} in {name}")
        if isinstance(d1[k], np.ndarray):
            if np.any(d2[k] != d1[k]):
                fatal(f"ERROR np array {k} {d1[k]} in {name}")
        else:
            if d2[k] != d1[k]:
                fatal(f"ERROR value for {k} in {name}")
    for k in d2:
        if k not in d1:
            fatal(f"ERROR, additional key {k} in {name}")


def ensure_directory_exists(directory):
    p = Path(directory)
    p.mkdir(parents=True, exist_ok=True)


def delete_folder_contents(folder_path):
    if os.path.exists(folder_path):
        for filename in os.listdir(folder_path):
            file_path = os.path.join(folder_path, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                warning(f"Failed to delete {file_path}. Reason: {e}")


# units were previously loaded dynamically from G4
# g4_units = Box()
# for t in g4.G4UnitDefinition.GetUnitsTable():
#     for a in t.GetUnitsList():
#         g4_units[str(a.GetName())] = a.GetValue()
#         g4_units[str(a.GetSymbol())] = a.GetValue()

# This dictionary was created with devtools.print_g4units_dict_string()
# It should be updated if Geant4 (ever) changes the units.
# It is hard-coded, as opposed to the dynamic variant above,
# because the build process on readthedocs cannot call G4 functions
g4_units = Box(
    {
        "parsec": 3.0856775807e19,
        "pc": 3.0856775807e19,
        "kilometer": 1000000.0,
        "km": 1000000.0,
        "meter": 1000.0,
        "m": 1000.0,
        "centimeter": 10.0,
        "cm": 10.0,
        "millimeter": 1.0,
        "mm": 1.0,
        "micrometer": 0.001,
        "um": 0.001,
        "nanometer": 1.0000000000000002e-06,
        "nm": 1.0000000000000002e-06,
        "angstrom": 1.0000000000000001e-07,
        "Ang": 1.0000000000000001e-07,
        "fermi": 1e-12,
        "fm": 1e-12,
        "kilometer2": 1000000000000.0,
        "km2": 1000000000000.0,
        "meter2": 1000000.0,
        "m2": 1000000.0,
        "centimeter2": 100.0,
        "cm2": 100.0,
        "millimeter2": 1.0,
        "mm2": 1.0,
        "barn": 9.999999999999999e-23,
        "millibarn": 9.999999999999999e-26,
        "mbarn": 9.999999999999999e-26,
        "microbarn": 9.999999999999999e-29,
        "mubarn": 9.999999999999999e-29,
        "nanobarn": 1e-31,
        "nbarn": 1e-31,
        "picobarn": 1e-34,
        "pbarn": 1e-34,
        "kilometer3": 1e18,
        "km3": 1e18,
        "meter3": 1000000000.0,
        "m3": 1000000000.0,
        "centimeter3": 1000.0,
        "cm3": 1000.0,
        "millimeter3": 1.0,
        "mm3": 1.0,
        "liter": 1000000.0,
        "L": 1000000.0,
        "dL": 100000.0,
        "cL": 10000.0,
        "mL": 1000.0,
        "radian": 1.0,
        "rad": 1.0,
        "milliradian": 0.001,
        "mrad": 0.001,
        "degree": 0.017453292519943295,
        "deg": 0.017453292519943295,
        "steradian": 1.0,
        "sr": 1.0,
        "millisteradian": 0.001,
        "msr": 0.001,
        "second": 1000000000.0,
        "s": 1000000000.0,
        "millisecond": 1000000.0,
        "ms": 1000000.0,
        "microsecond": 1000.0,
        "us": 1000.0,
        "nanosecond": 1.0,
        "ns": 1.0,
        "picosecond": 0.001,
        "ps": 0.001,
        "minute": 60000000000.0,
        "min": 60000000000.0,
        "hour": 3600000000000.0,
        "h": 3600000000000.0,
        "day": 86400000000000.0,
        "d": 86400000000000.0,
        "year": 3.1536e16,
        "y": 3.1536e16,
        "hertz": 1e-09,
        "Hz": 1e-09,
        "kilohertz": 1.0000000000000002e-06,
        "kHz": 1.0000000000000002e-06,
        "megahertz": 0.001,
        "MHz": 0.001,
        "cm/ns": 10.0,
        "mm/ns": 1.0,
        "cm/us": 0.01,
        "km/s": 0.001,
        "cm/ms": 1e-05,
        "m/s": 1e-06,
        "cm/s": 1e-08,
        "mm/s": 1e-09,
        "eplus": 1.0,
        "e+": 1.0,
        "coulomb": 6.241509074460763e18,
        "C": 6.241509074460763e18,
        "electronvolt": 1e-06,
        "eV": 1e-06,
        "kiloelectronvolt": 0.001,
        "keV": 0.001,
        "megaelectronvolt": 1.0,
        "MeV": 1.0,
        "gigaelectronvolt": 1000.0,
        "GeV": 1000.0,
        "teraelectronvolt": 1000000.0,
        "TeV": 1000000.0,
        "petaelectronvolt": 1000000000.0,
        "PeV": 1000000000.0,
        "millielectronVolt": 1e-09,
        "meV": 1e-09,
        "joule": 6241509074460.763,
        "J": 6241509074460.763,
        "eV/c": 1e-06,
        "keV/c": 0.001,
        "MeV/c": 1.0,
        "GeV/c": 1000.0,
        "TeV/c": 1000000.0,
        "GeV/cm": 100.0,
        "MeV/cm": 0.1,
        "keV/cm": 0.0001,
        "eV/cm": 1e-07,
        "milligram": 6.241509074460762e18,
        "mg": 6.241509074460762e18,
        "gram": 6.241509074460762e21,
        "g": 6.241509074460762e21,
        "kilogram": 6.241509074460762e24,
        "kg": 6.241509074460762e24,
        "g/cm3": 6.241509074460762e18,
        "mg/cm3": 6241509074460762.0,
        "kg/m3": 6241509074460762.0,
        "g/cm2": 6.241509074460761e19,
        "mg/cm2": 6.2415090744607624e16,
        "kg/cm2": 6.2415090744607614e22,
        "cm2/g": 1.6021766340000004e-20,
        "eV*cm2/g": 1.6021766340000002e-26,
        " eV*cm2/g": 1.6021766340000002e-26,
        "keV*cm2/g": 1.6021766340000002e-23,
        "MeV*cm2/g": 1.6021766340000004e-20,
        "GeV*cm2/g": 1.6021766340000003e-17,
        "watt": 6241.509074460762,
        "W": 6241.509074460762,
        "newton": 6241509074.460763,
        "N": 6241509074.460763,
        "pascal": 6241.509074460763,
        "Pa": 6241.509074460763,
        "bar": 624150907.4460763,
        "atmosphere": 632420906.9697368,
        "atm": 632420906.9697368,
        "ampere": 6241509074.460763,
        "A": 6241509074.460763,
        "milliampere": 6241509.0744607635,
        "mA": 6241509.0744607635,
        "microampere": 6241.509074460762,
        "muA": 6241.509074460762,
        "nanoampere": 6.2415090744607635,
        "nA": 6.2415090744607635,
        "volt": 1e-06,
        "V": 1e-06,
        "kilovolt": 0.001,
        "kV": 0.001,
        "megavolt": 1.0,
        "MV": 1.0,
        "volt/m": 9.999999999999999e-10,
        "V/m": 9.999999999999999e-10,
        "kilovolt/m": 1e-06,
        "kV/m": 1e-06,
        "megavolt/m": 0.001,
        "MV/m": 0.001,
        "weber": 1000.0,
        "Wb": 1000.0,
        "tesla": 0.001,
        "T": 0.001,
        "kilogauss": 0.0001,
        "kG": 0.0001,
        "gauss": 1.0000000000000001e-07,
        "G": 1.0000000000000001e-07,
        "kelvin": 1.0,
        "K": 1.0,
        "mole": 1.0,
        "mol": 1.0,
        "g/mole": 6.241509074460762e21,
        "g/mol": 6.241509074460762e21,
        "becquerel": 1e-09,
        "Bq": 1e-09,
        "curie": 37.0,
        "Ci": 37.0,
        "gray": 1.0000000000000002e-12,
        "Gy": 1.0000000000000002e-12,
    }
)


# def g4_units(name: str) -> float:
#     table = g4.G4UnitDefinition.GetUnitsTable()
#     for t in table:
#         for a in t.GetUnitsList():
#             if a.GetName() == name or a.GetSymbol() == name:
#                 return a.GetValue()
#     units_list = []
#     for t in table:
#         for a in t.GetUnitsList():
#             units_list.append(a.GetSymbol())
#     s = [str(u) + " " for u in units_list]
#     fatal(f"Error, cannot find the unit named {name}. Known are: {s}")
def get_material_name_variants(material_name):
    """Get different variants of a material name, e.g. with/without prepended G4_, only first letter capital.
    Intended to bridge inconsistencies in naming conventions.
    """
    # ensure the input is string, not G4String
    material_name = str(material_name)
    variants = [
        material_name,
        material_name.lstrip("G4_"),
        material_name.lstrip("G4_").capitalize(),
    ]
    return list(set(variants))


def g4_best_unit(value, unit_type):
    return g4.G4BestUnit(value, unit_type)


def g4_best_unit_tuple(value, unit_type):
    bu = g4.G4BestUnit(value, unit_type)
    parts = str(bu).split(" ", 1)
    float_part = float(parts[0])
    unit_part = parts[1].strip()
    return float_part, unit_part


def assert_key(key: str, d: Box):
    if key not in d:
        fatal(f'The key "{key}" is needed in this structure:\n' f"{d}")


def assert_keys(keys: list, d: Box):
    for key in keys:
        assert_key(key, d)


def indent(amount, text, ch=" "):
    """
    Prefix the text with indent spaces
    https://stackoverflow.com/questions/8234274/how-to-indent-the-contents-of-a-multi-line-string
    """
    return textwrap.indent(text, amount * ch)


def assert_unique_element_name(elements, name):
    if name in elements:
        s = (
            f"Error, cannot add '{name}' because this element's name already exists"
            f" in: {elements}."
        )
        fatal(s)


def make_builders(class_names):
    """
    Consider a list of Classname. For each, it builds a key/value, with:
    - the type of the class as key
    - and a lambda function that create an object of this class as value
    """
    builder_list = {}
    for c in class_names:
        # note the following lambda:
        # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
        try:
            builder_list[c.type_name] = lambda x, y=c: y(x)
        except AttributeError:
            # if type_name is not an attribute of the class,
            # we use the name of the class as key.
            # Also: no name parameter (this is for Physics List)
            builder_list[c.__name__] = lambda y=c: y()
    return builder_list


def read_mac_file_to_commands(filename):
    # read a file located into the 'mac' folder of the source code
    # return a list of commands
    resource_package = __package__
    with resources.open_text(f"{resource_package}.mac", filename) as f:
        c = f.read()
        commands = []
        for s in c.split("\n"):
            if s == "":
                continue
            # if s[0] == '#':
            #    continue
            commands.append(s)
        return commands


def ensure_filename_is_str(filename):
    # Some software packages, e.g. itk, do not support Path -> convert to str
    if isinstance(filename, Path):
        return str(filename)
    elif filename is None:
        return ""
    else:
        return filename


def insert_suffix_before_extension(file_path, suffix, suffix_separator="-"):
    path = Path(file_path)
    if not suffix:
        return path

    suffix = suffix.strip("_- *").lower()
    # Handle filenames with nested extensions e.g. '.nii.gz'
    if path.name.endswith(".nii.gz"):
        stem = path.name[: -len(".nii.gz")]
        new_path = path.with_name(f"{stem}{suffix_separator}{suffix}.nii.gz")
    else:
        new_path = path.with_name(f"{path.stem}{suffix_separator}{suffix}{path.suffix}")

    return new_path


def get_random_folder_name(size=8, create=True):
    r = "".join(random.choices(string.ascii_lowercase + string.digits, k=size))
    r = "run." + r
    directory = Path(r)
    if create:
        if not directory.exists():
            print(f"Creating output folder {r}")
            directory.mkdir(parents=True, exist_ok=True)
        if not directory.is_dir():
            fatal(f"Error, while creating {r}.")
    return r


def get_rnd_seed(seed):
    return RandomState(MT19937(SeedSequence(seed)))


def DDF():
    """
    Debug print current Function name
    """
    print("--> Entering", inspect.stack()[1][3])


def DD(arg):
    """
    Debug print variable name and its value
    """
    frame = inspect.currentframe()
    try:
        context = inspect.getframeinfo(frame.f_back).code_context
        caller_lines = "".join([line.strip() for line in context])
        m = re.search(r"DD\s*\((.+?)\);*$", caller_lines)
        if m:
            caller_lines = m.group(1)
            # end if
        print(caller_lines, "=", arg)
    finally:
        del frame


def print_dic(dic):
    print(json.dumps(dic, indent=4, default=str))


def get_release_date(opengate_version):
    import requests

    package_name = "opengate"
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)

    if response.status_code == 200:
        package_data = response.json()
        releases = package_data["releases"]
        if releases:
            # latest_version = max(releases, key=lambda v: package_data['releases'][v][0]['upload_time'])
            release_date = package_data["releases"][opengate_version][0]["upload_time"]
            return f"{release_date}"
        else:
            return "unknown"
    else:
        return "unknown"


def get_gate_folder():
    module_path = os.path.dirname(__file__)
    return Path(module_path)


def get_data_folder():
    return get_gate_folder() / "data"


def get_tests_folder():
    return get_gate_folder() / "tests" / "src"


def get_library_path():
    folder_base = (get_gate_folder() / "..").resolve()

    subpaths = [Path("opengate_core"), Path("core") / "opengate_core"]
    path = None

    for subpath in subpaths:
        try_path = folder_base / subpath
        if os.path.exists(try_path):
            path = try_path
            break

    if not path:
        return "unknown"

    files = os.listdir(path)
    lib_ext = "pyd" if os.name == "nt" else "so"
    libs = [file for file in files if file.endswith(f".{lib_ext}")]
    if len(libs) == 0:
        return "unknown"
    elif len(libs) > 1:
        return f"unknown: multiple .{lib_ext} files in {path} ({libs})"

    return path / libs[0]


def get_contrib_path():
    module_path = os.path.dirname(__file__)
    return Path(module_path) / "contrib"


def print_opengate_info():
    """
    Print information about OpenGate and the environment
    """

    gi = g4.GateInfo
    v = gi.get_G4Version().replace("$Name: ", "")
    v = v.replace("$", "")
    module_path = get_gate_folder()

    pv = sys.version.replace("\n", "")
    print(f"Python version   {pv}")
    print(f"Platform         {sys.platform}")
    print(f"Site package     {get_site_packages_dir()}")

    print(f"Geant4 version   {v}")
    print(f"Geant4 MT        {gi.get_G4MULTITHREADED()}")
    print(f"Geant4 Qt        {gi.get_G4VIS_USE_OPENGLQT()} {gi.get_QT_VERSION()}")
    print(f"Geant4 GDML      {gi.get_G4GDML()}")
    print(f"Geant4 date      {gi.get_G4Date().replace(')', '').replace('(', '')}")
    print(f"Geant4 data      {g4.get_g4_data_folder()}")

    print(f"ITK version      {gi.get_ITKVersion()}")

    print(f"GATE version     {version('opengate')}")
    print(f"GATE folder      {module_path}")
    print(f"GATE data        {get_data_folder()}")
    print(f"GATE tests       {get_tests_folder()}")
    print(f"GATE core path   {get_library_path()}")

    # check if from a git version ?
    git_path = Path(module_path).parent
    try:
        git_repo = git.Repo(git_path)
        sha = git_repo.head.object.hexsha
        print(f"GATE git sha     {sha}")
        commit = git_repo.head.commit
        commit_date = commit.committed_datetime
        print(f"GATE date        {commit_date}  (last commit)")

    except:
        print(f"GATE date        {get_release_date(version('opengate'))} (pypi)")


def calculate_variance(value_array, squared_value_array, number_of_samples):
    return np.clip(
        (
            squared_value_array / number_of_samples
            - np.power(value_array / number_of_samples, 2)
        )
        / (number_of_samples - 1),
        0,
        None,
    )


def standard_error_c4_correction(n):
    """
    Parameters
    ----------
    n : integer
        Number of subsets (of the samples).

    Returns
    -------
    c4 : double
        Factor to convert the biased standard error of the mean of subsets of the sample into an unbiased
        -  assuming a normal distribution .
        Usage: standard_error(unbiased) = standard_deviation_of_mean(=biased) / c4
        The reason is that the standard deviation of the mean of subsets of the sample X underestimates the true standard error. For n = 2 this underestimation is about 25%.

        Values for c4: n=2: 0.7979; n= 9: 0.9693

    """
    return (
        np.sqrt(2 / (n - 1)) * sc.special.gamma(n / 2) / sc.special.gamma((n - 1) / 2)
    )


def read_json_file(filename: Path) -> dict:
    """
    Read a JSON file into a Python dictionary.

    :param filename: Path object
        The filename of the JSON file to read.
    :return: dict
        The data from the JSON file.
    """
    if not filename.is_file():
        fatal(f"File {filename} does not exist.")

    with open(filename, "rb") as f:
        return json.load(f)


def get_basename_and_extension(filename):
    """Return the basename and extension of a filename even if .nii.gz is used."""
    base = filename
    extensions = []
    while os.path.splitext(base)[1]:
        base, ext = os.path.splitext(base)
        extensions.append(ext)
    extensions.reverse()
    return os.path.basename(base), "".join(extensions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.linacs.elektaversa as versa
import opengate.contrib.linacs.dicomrtplan as rtplan
from opengate.tests import utility
from opengate.utility import g4_units
from scipy.spatial.transform import Rotation
import numpy as np
import itk


def calc_mlc_aperture(x_leaf_position, y_jaws_position):

    left = x_leaf_position[:80]
    right = x_leaf_position[80:]
    proj_leaves_y = np.arange(-40 * 5 + 2.5, 40 * 5 - 2.5 + 0.1, 5)

    left[
        (proj_leaves_y - 2.5 < y_jaws_position[0])
        | (proj_leaves_y + 2.5 > y_jaws_position[1])
    ] = 0
    right[
        (proj_leaves_y - 2.5 < y_jaws_position[0])
        | (proj_leaves_y + 2.5 > y_jaws_position[1])
    ] = 0
    diff = np.array(np.abs(right - left)) * 5

    return np.sum(diff)


def add_volume_to_irradiate(sim, name, l_cp):
    mm = g4_units.mm
    m = g4_units.m
    plane = sim.add_volume("Box", "water_plane")
    plane.mother = name
    plane.material = "G4_WATER"
    plane.size = [0.4 * m, 0.4 * m, 2 * cm]
    plane.translation = [0 * mm, 0 * mm, -1 * cm]
    plane.color = [1, 0, 0, 1]  # red

    voxel_size_x = 1.2 * mm
    voxel_size_y = 1.2 * mm
    voxel_size_z = 2 * cm

    dim_box = [
        plane.size[0] / voxel_size_x,
        plane.size[1] / voxel_size_y,
        1,
    ]
    dose = sim.add_actor("DoseActor", "dose_water_slice")
    dose.attached_to = plane
    dose.edep.output_filename = "dose_actor_versa_rt_plan.mhd"  # FIXME
    dose.size = [int(dim_box[0]), int(dim_box[1]), int(dim_box[2])]
    dose.spacing = [voxel_size_x, voxel_size_y, voxel_size_z]
    dose.hit_type = "random"

    # move the plane
    rotations = []
    translations = []
    rotation_angle = rt_plan_parameters["gantry angle"]
    for n in l_cp:
        rot = Rotation.from_euler("y", rotation_angle[n], degrees=True)
        rot = rot.as_matrix()
        rotations.append(rot)
        translations.append(np.zeros(3))
    plane.add_dynamic_parametrisation(rotation=rotations, translation=translations)


def add_alpha_source(sim, name, nb_part):
    linac = sim.volume_manager.get_volume(name)
    z_linac = linac.size[2]
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    plan_source = sim.add_volume("Box", "plan_alpha_source")
    plan_source.material = "G4_Galactic"
    plan_source.mother = name
    plan_size = np.array([1 * nm, 1 * nm, 1 * nm])
    plan_source.size = np.copy(plan_size)
    plan_source.translation = [0 * mm, 0 * mm, z_linac / 2 - 3.5 * mm]

    source = sim.add_source("GenericSource", "alpha_source")
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    source.particle = "alpha"
    source.attached_to = plan_source.name
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV
    source.position.type = "box"
    source.position.size = np.copy(plan_size)
    source.direction.type = "iso"
    source.direction_relative_to_attached_volume = True
    source.direction.momentum = [0, 0, -1]
    source.activity = nb_part * Bq / sim.number_of_threads
    source.direction.theta = [0, 17 * deg]
    source.direction.phi = [0, 360 * deg]


def validation_test_19_rt_plan(
    theoretical_calculation,
    MC_calculation,
    cp_id,
    rt_plan_parameters,
    nb_part_init,
    nb_part_sent,
    tol=0.3,
):
    print(
        "Area from theoretical calculations for the chosen CP:",
        theoretical_calculation,
        "mm2",
    )
    print("Area from MC simulations for the chosen CP:", MC_calculation, "mm2")

    print("Number of particles emitted:", nb_part_sent)
    percentage_diff = (
        100 * (theoretical_calculation - MC_calculation) / theoretical_calculation
    )
    bool_percentage_diff = np.abs(percentage_diff) > tol * 100
    monitor_units = rt_plan_parameters["weight"][cp_id]
    nb_part_theo = nb_part_init * monitor_units
    print("Number of particles theoretically emitted:", int(np.round(nb_part_theo)))
    err_nb_part = np.sqrt(nb_part_theo)
    if (
        (nb_part_sent >= nb_part_theo - 4 * err_nb_part)
        and (nb_part_sent <= nb_part_theo + 4 * err_nb_part)
        and np.sum(bool_percentage_diff) == 0
    ):
        return True
    else:
        print("")
        print("FAIL")
        print(f"mm2 -> {percentage_diff=} (tol={tol}")
        print(f"{np.sum(bool_percentage_diff)=}")
        print(f"{nb_part_sent/nb_part_theo=}")
        print(f"{err_nb_part=}")
        return False


if __name__ == "__main__":
    # paths
    paths = utility.get_default_test_paths(__file__, output_folder="test019_linac")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "qt"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.random_seed = 123456789
    # sim.random_seed = "auto"
    sim.check_volumes_overlap = True

    # unit
    nm = gate.g4_units.nm
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    sec = gate.g4_units.s
    Bq = gate.g4_units.Bq

    # world
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_Galactic"
    a = np.array([0])

    # linac
    versa.add_linac_materials(sim)
    sad = 1000 * mm
    linac = versa.add_empty_linac_box(sim, "linac_box", sad)
    linac.material = "G4_Galactic"

    # jaws

    jaws = versa.add_jaws(sim, linac.name)

    # mlc
    mlc = versa.add_mlc(sim, linac.name)
    mlc_box = sim.volume_manager.get_volume(f"linac_box_mlc")
    mlc_box.material = "G4_Galactic"
    # add alpha source :
    nb_part = 1000000
    z_linac = linac.size[2]
    rt_plan_parameters = rtplan.read(str(paths.data / "DICOM_RT_plan.dcm"))
    MU = 0
    while MU == 0:
        l_cp = [np.random.randint(0, len(rt_plan_parameters["jaws 1"]), 1)[0]]
        # l_cp = [3]
        print("Control point: ", l_cp)
        MU = rt_plan_parameters["weight"][l_cp[0]]
    nb_part = nb_part / MU
    versa.set_linac_head_motion(
        sim, linac.name, jaws, mlc, rt_plan_parameters, sad=sad, cp_id=l_cp
    )

    if sim.visu:
        add_alpha_source(sim, linac.name, 100)
    else:
        add_alpha_source(sim, linac.name, nb_part)

    # Linac head rotation and jaws and mlc translation according to a DICOM rt plan

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 1000 * m)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # add water slice with a dose actor and a motion actor
    add_volume_to_irradiate(sim, world.name, l_cp)

    # start simulation
    # The number of particles provided (sim.activity) will be adapted
    # according to the number of MU delivered at each control points.
    versa.set_time_intervals_from_rtplan(sim, rt_plan_parameters, cp_id=l_cp)
    # leaves_position = rt_plan_parameters["leaves"]
    # jaw_1_positions = rt_plan_parameters["jaws 1"]
    # jaw_2_positions = rt_plan_parameters["jaws 2"]

    leaves = rt_plan_parameters["leaves"][l_cp[0]]
    jaws_1 = rt_plan_parameters["jaws 1"][l_cp[0]]
    jaws_2 = rt_plan_parameters["jaws 2"][l_cp[0]]
    jaws = [jaws_1, jaws_2]
    theoretical_area = calc_mlc_aperture(leaves, jaws)

    sim.run()

    # print results
    print(stats)

    # test

    dose2 = sim.get_actor("dose_water_slice")
    img_MC = dose2.edep.image
    # img_MC = itk.imread(dose2.get_output_path("edep"))
    array_MC = itk.GetArrayFromImage(img_MC)
    bool_MC = array_MC[array_MC != 0]
    simulated_area = len(bool_MC) * 1.44
    is_ok = validation_test_19_rt_plan(
        theoretical_area,
        simulated_area,
        l_cp[0],
        rt_plan_parameters,
        nb_part,
        stats.counts.events,
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import json
from box import Box
from opengate.contrib.dose.doserate import create_simulation
from opengate.utility import get_random_folder_name

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("json_param", nargs=1)
@click.option("--output_folder", "-o", default="AUTO", help="output folder, auto=rnd")
def go(json_param, output_folder):
    # open the param file
    try:
        f = open(json_param, "r")
        param = json.load(f)
    except IOError:
        print(f"Cannot open input json file {json_param}")
    param = Box(param)
    print(param)

    # set or create output_folder
    if output_folder == "AUTO":
        output_folder = get_random_folder_name()
    param.output_folder = output_folder

    # set activity as int (to deal with 1e4 notation)
    param.activity_bq = int(float(param.activity_bq))

    # create the simu
    sim = create_simulation(param)

    # run
    sim.run()

    # print results at the end
    stats = sim.get_actor("Stats")
    print(stats)
    print(f"Output in {param.output_folder}")


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 16 15:00:21 2024

@author: fava
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility
import itk
import numpy as np


def calculate_mean_unc(edep_arr, unc_arr, edep_thresh_rel=0.7):
    edep_max = np.amax(edep_arr)
    mask = edep_arr > edep_max * edep_thresh_rel
    unc_used = unc_arr[mask]
    unc_mean = np.mean(unc_used)

    return unc_mean


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test066"
    )

    # check statistical uncertainty every n_check simulated particles
    n_planned = 650000
    n_threads = 3

    # goal uncertainty
    unc_goal = 0.05  # means 5%
    thresh_voxel_edep_for_unc_calc = (
        0.7  # calculated over the voxels whose value is > 0.7 * max edep value
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 983456
    sim.number_of_threads = n_threads

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    um = gate.g4_units.um
    nm = gate.g4_units.nm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.set_production_cut("world", "all", 700 * um)

    # default source for tests
    # the source is fixed at the center, only the volume will move
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 90 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = n_planned * Bq  # 1 part/s

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test066-edep.mhd"
    dose.attached_to = "waterbox"
    dose.size = [40, 40, 40]
    mm = gate.g4_units.mm
    dose.spacing = [2.5 * mm, 2.5 * mm, 2.5 * mm]
    dose.edep_uncertainty.active = True
    dose.uncertainty_goal = unc_goal
    dose.uncertainty_first_check_after_n_events = 100
    dose.uncertainty_voxel_edep_threshold = thresh_voxel_edep_for_unc_calc
    dose.write_to_disk = False

    # add a stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True
    stat.write_to_disk = False

    # start simulation
    sim.run()

    # print results at the end
    print(stat)

    # test that final mean uncertainty satisfies the goal uncertainty
    test_thresh_rel = 0.01

    edep_arr = np.asarray(dose.edep.image)
    unc_array = np.asarray(dose.edep_uncertainty.image)

    unc_mean = calculate_mean_unc(
        edep_arr, unc_array, edep_thresh_rel=thresh_voxel_edep_for_unc_calc
    )
    print(f"{unc_goal = }")
    print(f"{unc_mean = }")
    ok = unc_mean < unc_goal * 1.01 and unc_mean > unc_goal - test_thresh_rel

    # test that the simulation stopped because of the threshold crtierion,
    # and not simply because we reached the planned number of events
    n_planned = n_planned * n_threads
    n_effective = stat.counts.events
    print(f"{n_planned = }")
    print(f"{n_effective = }")
    ok = ok and n_effective < n_planned

    utility.test_ok(ok)



import re
import os
from typing import List

import numpy as np
import itk
import json
from anytree import NodeMixin
from scipy.spatial.transform import Rotation

import opengate_core as g4

from ..base import DynamicGateObject, process_cls
from . import solids
from ..utility import ensure_filename_is_str
from ..exception import fatal, warning
from ..image import write_itk_image
from ..image import update_image_py_to_cpp
from .utility import (
    vec_np_as_g4,
    rot_np_as_g4,
    ensure_is_g4_transform,
)
from ..decorators import requires_fatal, requires_attribute_fatal
from ..definitions import __world_name__, __gate_list_objects__
from ..actors.dynamicactors import (
    VolumeImageChanger,
    VolumeTranslationChanger,
    VolumeRotationChanger,
)
from .materials import create_density_img, write_material_database
from opengate.serialization import dump_json
from opengate.utility import g4_units


def _setter_hook_user_info_rotation(self, rotation_user):
    """Internal function associated with user_info rotation to check its validity."""
    if rotation_user is None:
        rotation = [Rotation.identity().as_matrix()]
    elif isinstance(rotation_user, (np.matrix, np.ndarray)) and rotation_user.shape == (
        3,
        3,
    ):
        # user has provided a single rotation matrix
        rotation = [rotation_user]
    elif isinstance(rotation_user[0], (np.matrix, np.ndarray)) and all(
        [
            isinstance(r, (np.matrix, np.ndarray)) and r.shape == (3, 3)
            for r in rotation_user
        ]
    ):
        # user has provided a list of rotation matrices
        rotation = rotation_user
    else:
        fatal(
            f"The parameter 'rotation' should be a 3x3 array or matrix or a list of such arrays/matrices. "
            f"You provided {rotation_user} for volume {self.name}. "
        )
    return rotation


def _getter_hook_user_info_rotation(self, rotation):
    if len(rotation) == 1:
        return rotation[0]
    else:
        return rotation


def _setter_hook_user_info_translation(self, translation_user):
    # if the user passes a single 3-vector, its first entry will be a number
    # ensure that translation is a list of vectors
    if translation_user is None:
        translation = [np.zeros(3, dtype=float)]
    elif not isinstance(translation_user[0], (__gate_list_objects__, np.ndarray)):
        translation = [np.array(translation_user)]
    else:
        translation = np.array(translation_user)
    if not all([len(t) == 3 for t in translation]):
        fatal(
            f"The translation parameter must be a 3-vector or a list of 3-vectors, "
            f"e.g. [1,2,1] or [[2,4,3], [5,4,7]]. "
            f"For volume {self.name}, you provided: \n{translation_user} for volume {self.name}. "
        )
    return translation


def _getter_hook_user_info_translation(self, translation):
    if len(translation) == 1:
        return translation[0]
    else:
        return translation


def _setter_hook_user_info_mother(self, mother):
    """Hook to be attached to property setter of user info 'mother' in all volumes.

    Checks if new mother is actually different from stored one.\n
    If so, it also tries to inform the volume manager that the volume tree needs an update. \n
    This latter part only applies for volumes which have a volume manager, \n
    i.e. which have been added to a simulation.
    """
    # duck typing: allow volume objects or their name
    try:
        mother_name = mother.name
    except AttributeError:
        mother_name = mother
    if mother_name != self.user_info["mother"]:
        try:
            self.volume_manager._need_tree_update = True
        except AttributeError:
            pass
    return mother_name


def _setter_hook_voxel_materials(self, voxel_materials):
    # make a structured array (https://stackoverflow.com/a/44337962)
    vm = np.array(
        [tuple(row) for row in voxel_materials],
        dtype=np.dtype(
            {"names": ["lower", "upper", "material"], "formats": [float, float, "<U32"]}
        ),
    )
    np.sort(vm, order="lower")
    return vm


def _setter_hook_ensure_array(self, input):
    return np.asarray(input)  # becomes dtype='<U32'


# inherit from NodeMixin to turn the class into a tree node
class VolumeBase(DynamicGateObject, NodeMixin):
    """
    Store information about a geometry volume:
    - G4 objects: Solid, LogicalVolume, PhysicalVolume
    - user parameters: user_info
    - additional data such as: mother, material etc
    """

    user_info_defaults = {
        "mother": (
            __world_name__,
            {
                "doc": "Name of the mother volume.",
                "setter_hook": _setter_hook_user_info_mother,
            },
        ),
        "material": ("G4_AIR", {"doc": "Name of the material"}),
        "translation": (
            [[0, 0, 0]],
            {
                "doc": "3-component vector or list of such vectors defining the translation "
                "w.r.t. the mother. If translation is a list of vectors, "
                "the volume will be repeeted once for each translation vector.",
                "setter_hook": _setter_hook_user_info_translation,
                "getter_hook": _getter_hook_user_info_translation,
                "dynamic": True,
            },
        ),
        "color": (
            [1, 1, 1, 1],
            {
                "doc": (
                    "4 component vector defining the volume's color in visual rendering. "
                    "The first 3 entries are RBG, the 4th is visible/invisible (1 or 0). "
                )
            },
        ),
        "rotation": (
            [Rotation.identity().as_matrix()],
            {
                "doc": "3x3 rotation matrix or list of such matrices. "
                "The matrix (matrices) should be np.array or np.matrix."
                "If a list of matrices is provided, the volume will be repeated, "
                "once for each rotation vector.",
                "setter_hook": _setter_hook_user_info_rotation,
                "getter_hook": _getter_hook_user_info_rotation,
                "dynamic": True,
            },
        ),
        "build_physical_volume": (
            True,
            {
                "doc": "Boolean flag (True/False) whether G4 should build a physical volume.",
                "type": bool,
            },
        ),
    }

    def __init__(self, *args, template=None, **kwargs):
        # GateObject base class digests all user info provided as kwargs
        # template = kwargs.pop('template', None)
        super().__init__(*args, **kwargs)

        # if a template volume is provided, clone all user info items from it
        # except for the name of course
        if template is not None:
            # FIXME: consider using from_dictionary()
            self.configure_like(template)
            # put back user infos which were explicitly passed as keyword argument
            for k in self.user_info.keys():
                if k != "name" and k in kwargs:
                    setattr(self, k, kwargs[k])

        # this attribute is used internally for the volumes tree
        # do not set it manually!
        self.parent = None

        self._is_constructed = False
        self.volume_engine = None

        # G4 references
        self.g4_world_log_vol = None
        self.g4_logical_volume = None
        self.g4_vis_attributes = None
        # this list contains all physical volumes (in case of repeated volume)
        self.g4_physical_volumes = []
        self.g4_material = None

    def close(self):
        self.release_g4_references()
        self.volume_engine = None
        self._is_constructed = False
        super().close()

    def release_g4_references(self):
        self.g4_logical_volume = None
        self.g4_vis_attributes = None
        self.g4_physical_volumes = []
        self.g4_material = None

    def __getstate__(self):
        return_dict = super().__getstate__()
        # Reset the following references to None because they cannot be pickled
        # They created when running a simulation
        return_dict["g4_logical_volume"] = None
        return_dict["g4_vis_attributes"] = None
        return_dict["g4_physical_volumes"] = []
        return_dict["g4_material"] = None
        return_dict["volume_engine"] = None
        return_dict["_is_constructed"] = False
        return return_dict

    def __finalize_init__(self):
        super().__finalize_init__()
        # need to add this explicitly because anytree does not properly declare
        # the attribute __parent in the NodeMixin.__init__ which leads to falls warnings
        self.known_attributes.add("_NodeMixin__parent")
        self.known_attributes.add("_NodeMixin__children")

    def _update_node(self):
        """Internal method which retrieves the volume object
        from the volume manager based on the mother's name stored as user info 'mother'
        """
        try:
            self.parent = self.volume_manager.get_volume(self.mother)
        except KeyError:
            fatal(
                "Error while trying to update a volume tree node: \n"
                f"Mother volume of {self.name} should be {self.mother}, "
                f"but it cannot be found in the list of volumes in the volume manager."
                f"Known volumes are: \n{self.volume_manager.volumes}"
            )

    def _request_volume_tree_update(self):
        try:
            self.volume_manager.update_volume_tree()
        except AttributeError:
            fatal(
                f"Unable to determine the world volume to which volume named {self.name} belongs. "
                "Probably the volume has not yet been added to the simulation. "
            )

    # # FIXME: maybe store reference to simulation directly, rather than reference to volume_manager?
    @property
    def volume_manager(self):
        # It is not compulsory for a GateObject to belong to a simulation,
        # and the volume might therefore not have any reference to a volume manager
        # (e.g. in volumes created for boolean operations),
        # but without a simulation/volume_manager, the volume is not known to the simulation
        # and certain functionality is unavailable
        if self.simulation is not None:
            return self.simulation.volume_manager
        else:
            return None

    @property
    def volume_type(self):
        return type(self).__name__

    @property
    def world_volume(self):
        self._request_volume_tree_update()
        try:
            self.volume_manager.update_volume_tree()
        except AttributeError:
            fatal(
                f"Unable to determine the world volume to which volume named {self.name} belongs. "
                "Probably the volume has not yet been added to the simulation. "
            )
        try:
            return self.ancestors[
                1
            ]  # index 0 is the volume tree root, index 1 is world level
        except IndexError:  # if no ancestors, this is a world volume already
            return self

    @property
    def volume_depth_in_tree(self):
        self._request_volume_tree_update()
        return len(self.ancestors) - 1  # do not count the tree root

    @property
    def ancestor_volumes(self):
        self._request_volume_tree_update()
        return self.ancestors[1:]  # first item is volume tree root, not a real volume

    @property
    def children_volumes(self):
        self._request_volume_tree_update()
        return self.children

    @property
    def number_of_repetitions(self):
        return len(self.user_info["translation"])

    def get_g4_physical_volume(self, index):
        try:
            return self.g4_physical_volumes[index]
        except IndexError:
            fatal(
                f"No physical volume with repetition index {index} "
                f"found in volume {self.name}. "
            )

    @property
    def translation_list(self):
        """Utility property which always returns a list of translations,
        even if the volume is not repeated and has thus only one translation vector.
        """
        len_rot = len(self.user_info["rotation"])
        if len_rot > 1 and len(self.user_info["translation"]) == 1:
            return self.user_info["translation"] * len_rot
        else:
            return self.user_info["translation"]

    @property
    def rotation_list(self):
        """Utility property which always returns a list of rotations,
        even if the volume is not repeated and has thus only one rotation vector.
        """
        len_trans = len(self.user_info["translation"])
        if len_trans > 1 and len(self.user_info["rotation"]) == 1:
            return self.user_info["rotation"] * len_trans
        else:
            return self.user_info["rotation"]

    @property
    def g4_region(self):
        if self.g4_logical_volume is None:
            return None
        else:
            return self.g4_logical_volume.GetRegion()

    # shortcut to first physical volume
    # FIXME: remove this shortcut. confusingly similar to self.g4_physical_volumes
    @property
    def g4_physical_volume(self):
        return self.g4_physical_volumes[0]

    # shortcuts to G4 variants of user info items 'translation' and 'rotation'
    @property
    def g4_translation(self):
        return [vec_np_as_g4(t) for t in self.translation_list]

    @property
    def g4_rotation(self):
        try:
            return [rot_np_as_g4(r) for r in self.rotation_list]
        except Exception as e:
            fatal(
                f"Unable to create G4 rotation matrix in volume {self.name}. "
                f"\nOriginal error message: {e}."
            )

    @property
    def g4_transform(self):
        g4_translation = self.g4_translation
        g4_rotation = self.g4_rotation
        if len(g4_translation) != len(g4_rotation):
            fatal(
                f"The number of translation vectors and rotation matrices in volume '{self.name}' does not match. "
                f"I found {len(g4_translation)} translations and {len(g4_rotation)} rotations. "
            )
        return [
            ensure_is_g4_transform(t, r) for t, r in zip(g4_translation, g4_rotation)
        ]

    # shortcut to the G4LogicalVolume of the mother
    @property
    def mother_g4_logical_volume(self):
        if self.mother is None:
            return None
        else:
            return g4.G4LogicalVolumeStore.GetInstance().GetVolume(self.mother, False)

    @property
    def mother_volume(self):
        self._update_node()
        return self.parent

    def construct(self):
        if self._is_constructed is False:
            self.construct_material()
            self.construct_solid()
            self.construct_logical_volume()
            # check user info:
            if self.build_physical_volume is True:
                self.construct_physical_volume()
            self._is_constructed = True

    def construct_material(self):
        if self.volume_manager is None:
            fatal(
                f"The volume {self.name} does not seem to be added to the simulation. "
                f"Use sim.volume_manager.add_volume(...) to add it. "
            )
        # retrieve or build the material
        if self.material is None:
            self.g4_material = None
        else:
            self.g4_material = self.volume_manager.find_or_build_material(self.material)

    @requires_attribute_fatal("g4_solid")
    @requires_fatal("g4_material")
    def construct_logical_volume(self):
        self.g4_logical_volume = g4.G4LogicalVolume(
            self.g4_solid, self.g4_material, self.name
        )
        # color
        self.g4_vis_attributes = g4.G4VisAttributes()
        self.g4_vis_attributes.SetColor(*self.color)
        self.g4_vis_attributes.SetVisibility(bool(self.color[3]))
        self.g4_logical_volume.SetVisAttributes(self.g4_vis_attributes)

    def construct_physical_volume(self):
        g4_transform = self.g4_transform
        if len(g4_transform) > 1:
            fatal(
                f"The volume named {self.name} of type {type(self).__name__} is not repeatable. "
                f"You may therefore only provide a single translation and/or rotation vector. "
            )
        self.g4_physical_volumes = [
            self._make_physical_volume(self.name, g4_transform[0])
        ]

    @requires_fatal("volume_manager")
    def _make_physical_volume(self, volume_name, g4_transform, copy_index=0):
        return g4.G4PVPlacement(
            g4_transform,
            self.g4_logical_volume,  # logical volume
            volume_name,  # volume name
            self.mother_g4_logical_volume,  # mother volume or None if World
            False,  # no boolean operation # FIXME for BooleanVolume ?
            copy_index,  # copy number
            self.volume_manager.simulation.check_volumes_overlap,
        )  # overlaps checking

    def create_changers(self):
        changers = super().create_changers()
        for dp in self.dynamic_params.values():
            if dp["extra_params"]["auto_changer"] is True:
                if "translation" in dp:
                    new_changer = VolumeTranslationChanger(
                        name=f"{self.name}_volume_translation_changer_{len(changers)}",
                        translations=dp["translation"],
                        attached_to=self,
                        simulation=self.simulation,
                        repetition_index=dp["extra_params"].pop("repetition_index", 0),
                    )
                    changers.append(new_changer)
                if "rotation" in dp:
                    new_changer = VolumeRotationChanger(
                        name=f"{self.name}_volume_rotation_changer_{len(changers)}",
                        rotations=dp["rotation"],
                        attached_to=self,
                        simulation=self.simulation,
                        repetition_index=dp["extra_params"].pop("repetition_index", 0),
                    )
                    changers.append(new_changer)
            else:
                warning(
                    f"You need to manually create a changer for dynamic parametrisation {dp} "
                    f"of volume '{self.name}'."
                )
        return changers

    # set physical properties in this (logical) volume
    # behind the scenes, this will create a region and associate this volume with it
    @requires_fatal("volume_manager")
    def set_production_cut(self, particle_name, value):
        self.volume_manager.simulation.physics_manager.set_production_cut(
            self.name, particle_name, value
        )

    @requires_fatal("volume_manager")
    def set_max_step_size(self, max_step_size):
        self.volume_manager.simulation.physics_manager.set_max_step_size(
            self.name, max_step_size
        )

    @requires_fatal("volume_manager")
    def set_max_track_length(self, max_track_length):
        self.volume_manager.simulation.physics_manager.set_max_track_length(
            self.name, max_track_length
        )

    @requires_fatal("volume_manager")
    def set_min_ekine(self, min_ekine):
        self.volume_manager.simulation.physics_manager.set_min_ekine(
            self.name, min_ekine
        )

    @requires_fatal("volume_manager")
    def set_max_time(self, max_time):
        self.volume_manager.simulation.physics_manager.set_max_time(self.name, max_time)

    @requires_fatal("volume_manager")
    def set_min_range(self, min_range):
        self.volume_manager.simulation.physics_manager.set_min_range(
            self.name, min_range
        )


class RepeatableVolume(VolumeBase):
    def get_repetition_name_from_index(self, index):
        return f"{self.name}_rep_{index}"

    def get_repetition_index_from_name(self, name):
        suffix = re.findall(r"_rep_\d+", name)
        if len(suffix) != 1:
            fatal(
                f"Something went wrong while trying to determine the repetition index "
                f"from repetition name {name} in volume {self.name}."
            )
        else:
            suffix = suffix[0]
        return int(suffix.lstrip("rep_"))

    def construct_physical_volume(self):
        g4_transform = self.g4_transform
        if len(g4_transform) > 1:
            self.g4_physical_volumes = []  # reset list to empty
            for i, g4t in enumerate(g4_transform):
                self.g4_physical_volumes.append(
                    self._make_physical_volume(
                        self.get_repetition_name_from_index(i),
                        g4t,
                        copy_index=i,
                    ),
                )
        else:
            super().construct_physical_volume()

    def add_dynamic_parametrisation(self, repetition_index=0, **params):
        super().add_dynamic_parametrisation(repetition_index=repetition_index, **params)


class BooleanVolume(RepeatableVolume, solids.BooleanSolid):
    """
    Volume resulting from a boolean operation
    of the solids contained in two volumes.
    """


# Function to handle boolean operations on volumes
# They create a new volume object, but the actual g4_solid is only created when the volume's
# construct() method is invoked
def _make_boolean_volume(
    volume_1, volume_2, operation, translation=None, rotation=None, new_name=None
):
    name_joiners = {"intersect": "times", "add": "plus", "subtract": "minus"}
    if operation not in name_joiners.keys():
        fatal("Unknown boolean operation. ")
    if new_name is None:
        new_name = f"({volume_1.name}_{name_joiners[operation]}_{volume_2.name})"

    if rotation is None:
        rotation = Rotation.identity().as_matrix()
    if translation is None:
        translation = [0, 0, 0]

    new_volume = BooleanVolume(name=new_name, template=volume_1)
    # need to access the user_info dict directly because the property 'creator_volumes' is read-only
    new_volume.user_info["creator_volumes"] = [volume_1, volume_2]
    new_volume.rotation_boolean_operation = rotation
    new_volume.translation_boolean_operation = translation
    new_volume.operation = operation
    return new_volume


def intersect_volumes(
    volume_1, volume_2, translation=None, rotation=None, new_name=None
):
    """Creates a boolean representing the intersection of two volumes."""

    return _make_boolean_volume(
        volume_1,
        volume_2,
        "intersect",
        translation=translation,
        rotation=rotation,
        new_name=new_name,
    )


def unite_volumes(volume_1, volume_2, translation=None, rotation=None, new_name=None):
    """Creates a boolean representing the union of two volumes."""

    return _make_boolean_volume(
        volume_1,
        volume_2,
        "add",
        translation=translation,
        rotation=rotation,
        new_name=new_name,
    )


def subtract_volumes(
    volume_1, volume_2, translation=None, rotation=None, new_name=None
):
    """Creates a boolean representing the first volume minus the overlapping portion of the second volume."""

    return _make_boolean_volume(
        volume_1,
        volume_2,
        "subtract",
        translation=translation,
        rotation=rotation,
        new_name=new_name,
    )


# **** Specific CSG volumes ****
# They are defined by inheriting from the corresponding solid class
# Nothing else needs to be implemented here.


class BoxVolume(RepeatableVolume, solids.BoxSolid):
    """
    Volume with a box shape.
    """


class HexagonVolume(RepeatableVolume, solids.HexagonSolid):
    """
    Volume with a hexagon shape.
    """


class ConsVolume(RepeatableVolume, solids.ConsSolid):
    """
    Volume with the shape of a cone or conical section.
    """


class PolyhedraVolume(RepeatableVolume, solids.PolyhedraSolid):
    """
    Volume with a polyhedral shape.
    """


class SphereVolume(RepeatableVolume, solids.SphereSolid):
    """
    Volume with a sphere or spherical shell shape.
    """


class EllipsoidVolume(RepeatableVolume, solids.EllipsoidSolid):
    """
    Volume with an ellipsoid shape.
    """


class TrapVolume(RepeatableVolume, solids.TrapSolid):
    """
    Volume with a generic trapezoidal shape.
    """


class TrdVolume(RepeatableVolume, solids.TrdSolid):
    """
    Volume with a symmetric trapezoidal shape.
    """


class TubsVolume(RepeatableVolume, solids.TubsSolid):
    """
    Volume with a tube or cylindrical section shape.
    """


class TesselatedVolume(RepeatableVolume, solids.TesselatedSolid):
    """
    Volume based on a mesh volume by reading a mesh file.
    """


class RepeatParametrisedVolume(VolumeBase):
    """
    Volume created from another volume via translations.
    """

    user_info_defaults = {
        "linear_repeat": (
            [1, 1, 1],
            {"doc": "FIXME"},
        ),
        "offset": (
            [0, 0, 0],
            {"doc": "3 component vector or list."},
        ),
        "offset_nb": (1, {"doc": "FIXME"}),
        "start": ("auto", {"doc": "FIXME"}),
    }

    def __init__(self, repeated_volume, *args, **kwargs):
        # FIXME: This should probably be a user_info
        self.repeated_volume = repeated_volume
        if "name" not in kwargs:
            kwargs["name"] = f"{repeated_volume.name}_param"
        kwargs["mother"] = repeated_volume.mother
        super().__init__(*args, **kwargs)
        if repeated_volume.build_physical_volume is True:
            repeated_volume.build_physical_volume = False
        self.g4_repeat_parametrisation = None

    def close(self):
        self.repeated_volume.close()
        super().close()

    def release_g4_references(self):
        super().release_g4_references()
        # FIXME: unsure. If not set to None, we get the following error:
        # "cannot pickle 'opengate_core.opengate_core.GateRepeatParameterisation' object"
        self.g4_repeat_parametrisation = None

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_repeat_parametrisation"] = None
        return return_dict

    def construct(self):
        if self._is_constructed is False:
            # construct the repeated volume,
            # it will not construct the phys volume because that was disabled in init()
            self.repeated_volume.construct()
            # construct the physical volume of this repeat parametrised volume
            self.construct_physical_volume()
            self._is_constructed = True

    def construct_physical_volume(self):
        # check if the mother is the world
        if self.mother_g4_logical_volume is None:
            fatal(f"The mother of {self.name} cannot be the world.")

        self.create_repeat_parametrisation()

        # number of copies
        n = (
            self.linear_repeat[0]
            * self.linear_repeat[1]
            * self.linear_repeat[2]
            * self.offset_nb
        )

        # (only daughter)
        # g4.EAxis.kUndefined => faster
        self.g4_physical_volumes.append(
            g4.G4PVParameterised(
                self.name,
                self.repeated_volume.g4_logical_volume,  # logical volume from the repeated volume
                self.mother_g4_logical_volume,
                g4.EAxis.kUndefined,
                n,
                self.g4_repeat_parametrisation,
                False,  # very slow if True
            )
        )

    def create_repeat_parametrisation(self):
        if self.start == "auto":
            self.start = [
                -(x - 1) * y / 2.0 for x, y in zip(self.linear_repeat, self.translation)
            ]
        # create parameterised
        keys = [
            "linear_repeat",
            "start",
            "translation",
            "rotation",
            "offset",
            "offset_nb",
        ]
        if self.number_of_repetitions > 1:
            fatal(
                f"The {type(self).name} volume named '{self.name}' has multiple translations/rotations, "
                f"but only one is allowed."
            )
        p = {}
        for k in keys:
            p[k] = getattr(self, k)
        self.g4_repeat_parametrisation = g4.GateRepeatParameterisation()
        self.g4_repeat_parametrisation.SetUserInfo(p)


def _setter_hook_image(self, image):
    if image != self.image:
        self._itk_image = None
    return image


class ImageVolume(VolumeBase, solids.ImageSolid):
    """
    Store information about a voxelized volume
    """

    voxel_materials: List
    image: str
    dump_label_image: str

    user_info_defaults = {
        "voxel_materials": (
            [[-np.inf, np.inf, "G4_AIR"]],
            {
                "doc": "FIXME",
            },
        ),
        "image": (
            "",
            {
                "doc": "Path to the image file",
                "is_input_file": True,
                "dynamic": True,
                "setter_hook": _setter_hook_image,
            },
        ),
        "dump_label_image": (
            None,
            {
                "doc": "Path at which the image containing material labels should be saved. "
                "Set to None to dump no image."
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # look up table MATERIAL -> LABEL
        self.material_to_label_lut = None

        # ITK images
        self._itk_image = None  # the input
        self.label_image = None  # image storing material labels
        # G4 references (additionally to those in base class)
        self.g4_physical_x = None
        self.g4_physical_y = None
        self.g4_physical_z = None
        self.g4_logical_x = None
        self.g4_logical_y = None
        self.g4_logical_z = None
        self.g4_voxel_param = None
        self.g4_vis_attributes_logical_slices = None

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_physical_x"] = None
        return_dict["g4_physical_y"] = None
        return_dict["g4_physical_z"] = None
        return_dict["g4_logical_x"] = None
        return_dict["g4_logical_y"] = None
        return_dict["g4_logical_z"] = None
        return_dict["g4_voxel_param"] = None
        return_dict["g4_vis_attributes_logical_slices"] = None
        return_dict["slice_xy"] = None
        return_dict["slice_xz"] = None
        return_dict["slice_yz"] = None
        return return_dict

    def close(self):
        self.release_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_logical_x = None
        self.g4_logical_y = None
        self.g4_logical_z = None
        self.g4_physical_x = None
        self.g4_physical_y = None
        self.g4_physical_z = None
        self.g4_voxel_param = None
        self.g4_vis_attributes_logical_slices = None
        self.slice_xy = None
        self.slice_xz = None
        self.slice_yz = None

    @property
    def itk_image(self):
        if self._itk_image is None:
            warning(
                f"The itk_image in {self.type_name} '{self.name}' is None. "
                f"If this is unexpected, run my_image_volume.load_input_image() first, "
                f"where my_image_volume is the variable name of the {self.type_name} in your script. "
            )
        return self._itk_image

    @itk_image.setter
    def itk_image(self, image):
        self._itk_image = image

    # @requires_fatal('itk_image')
    # FIXME: replace this property by function in opengate.image
    @property
    def size_pix(self):
        if self.itk_image is None:
            self.load_input_image()
        return np.array(itk.size(self.itk_image)).astype(int)

    # @requires_fatal('itk_image')
    # FIXME: replace this property by function in opengate.image
    @property
    def spacing(self):
        if self.itk_image is None:
            self.load_input_image()
        return np.array(self.itk_image.GetSpacing())

    # @requires_fatal("itk_image")
    @property
    def native_translation(self):
        if self.itk_image is not None:
            origin = np.array(self.itk_image.GetOrigin())
            spacing = np.array(self.itk_image.GetSpacing())
            size = np.array(self.itk_image.GetLargestPossibleRegion().GetSize())
            center = (size - 1.0) * spacing / 2.0
            return origin + Rotation.from_matrix(self.native_rotation).apply(center)
        else:
            return None

    # @requires_fatal("itk_image")
    @property
    def native_rotation(self):
        if self.itk_image is not None:
            return np.array(self.itk_image.GetDirection())
        else:
            return None

    def set_materials_from_voxelisation(self, labels_json):
        """
        Reads a JSON file containing label information and sets the voxel materials
        based on the data provided in the JSON file. The JSON file should contain a
        dictionary where each entry has a 'label' and 'material' field.

        The json file could have been generated with 'voxelize_geometry'

        Args:
            labels_json (str): Path to the JSON file containing label data.

        Raises:
            FileNotFoundError: If the specified JSON file cannot be found.
            json.JSONDecodeError: If the JSON file contains invalid JSON.

        Attributes:
            voxel_materials (list): A list where each entry is a list containing a
                                    label, the next label, and the corresponding material.
        """
        with open(labels_json) as f:
            labels = json.load(f)
        self.voxel_materials = [
            [l["label"], l["label"] + 1, l["material"]] for k, l in labels.items()
        ]

    @requires_fatal("volume_engine")
    def construct(self):
        self.material_to_label_lut = self.create_material_to_label_lut()
        # make sure the materials are created in Geant4
        for m in self.material_to_label_lut:
            self.volume_manager.find_or_build_material(m)
        self.itk_image = self.load_input_image()
        self.label_image = self.create_label_image()
        if self.dump_label_image:
            self.save_label_image()
        # set attributes of the solid
        self.half_size_mm = 0.5 * self.size_pix * self.spacing
        self.half_spacing = 0.5 * self.spacing
        self.construct_material()
        self.construct_solid()
        self.construct_slices()
        self.construct_logical_volume()
        self.g4_voxel_param = self.create_image_parametrisation()
        self.construct_physical_volume()

    def construct_slices(self):
        image_array = itk.GetArrayFromImage(self.itk_image)
        min_pixel = np.min(image_array)
        interval_pixel = np.max(image_array) - min_pixel
        self.slice_xy = (
            image_array[int(self.size_pix[2] / 2), :, :] - min_pixel
        ) / interval_pixel
        self.slice_xy = list(self.slice_xy.reshape(1, -1)[0])
        self.slice_xz = (
            image_array[:, int(self.size_pix[1] / 2), :] - min_pixel
        ) / interval_pixel
        self.slice_xz = list(self.slice_xz.reshape(1, -1)[0])
        self.slice_yz = (
            image_array[:, :, int(self.size_pix[0] / 2)] - min_pixel
        ) / interval_pixel
        self.slice_yz = list(self.slice_yz.reshape(1, -1)[0])
        image_dict = {
            "slice_xy": self.slice_xy,
            "slice_xz": self.slice_xz,
            "slice_yz": self.slice_yz,
        }
        self.g4_solid.SetSlices(image_dict)
        self.volume_manager.solid_with_texture_init.append(self.g4_solid)

    def construct_physical_volume(self):
        super().construct_physical_volume()

        self.g4_physical_y = g4.G4PVReplica(
            self.name + "_Y",
            self.g4_logical_y,
            self.g4_logical_volume,
            g4.EAxis.kYAxis,
            self.size_pix[1],  # nReplicas
            self.spacing[1],  # width
            0.0,
        )  # offset

        # param X
        self.g4_physical_x = g4.G4PVReplica(
            self.name + "_X",
            self.g4_logical_x,
            self.g4_logical_y,
            g4.EAxis.kXAxis,
            self.size_pix[0],
            self.spacing[0],
            0.0,
        )

        self.g4_physical_z = g4.G4PVParameterised(
            self.name + "_Z",
            self.g4_logical_z,
            self.g4_logical_x,
            g4.EAxis.kZAxis,  # g4.EAxis.kUndefined, ## FIXME ?
            self.size_pix[2],
            self.g4_voxel_param,
            False,
        )  # overlaps checking

    def construct_logical_volume(self):
        super().construct_logical_volume()
        self.g4_logical_x = g4.G4LogicalVolume(
            self.g4_solid_x, self.g4_material, self.name + "_log_X"
        )
        self.g4_logical_y = g4.G4LogicalVolume(
            self.g4_solid_y, self.g4_material, self.name + "_log_Y"
        )
        self.g4_logical_z = g4.G4LogicalVolume(
            self.g4_solid_z, self.g4_material, self.name + "_log_Z"
        )
        self.g4_vis_attributes_logical_slices = g4.G4VisAttributes()
        self.g4_vis_attributes_logical_slices.SetVisibility(bool(False))
        self.g4_logical_x.SetVisAttributes(self.g4_vis_attributes_logical_slices)
        self.g4_logical_y.SetVisAttributes(self.g4_vis_attributes_logical_slices)
        self.g4_logical_z.SetVisAttributes(self.g4_vis_attributes_logical_slices)

    def create_material_to_label_lut(self, material=None, voxel_materials=None):
        if voxel_materials is None:
            voxel_materials = self.voxel_materials
        if material is None:
            material = self.material
        # prepare a LUT from material name to label
        material_to_label_lut = {}
        material_to_label_lut[material] = 0  # initialize with label 0

        # sort voxel_materials according to lower bounds
        voxel_materials_sorted = sorted(voxel_materials, key=lambda x: x[0])

        lower_bounds = np.array([row[0] for row in voxel_materials_sorted])
        upper_bounds = np.array([row[1] for row in voxel_materials_sorted])
        if not (lower_bounds[1:] >= upper_bounds[:-1]).all():
            fatal(f"Overlapping intervals in voxel_materials of volume {self.name}.")

        # fill the LUT
        i = 1
        for row in voxel_materials_sorted:
            if row[2] not in material_to_label_lut:
                material_to_label_lut[row[2]] = i
                i += 1

        return material_to_label_lut

    def load_input_image(self, path=None):
        if path is None:
            itk_image = itk.imread(ensure_filename_is_str(self.image))
            self.itk_image = itk_image
        else:
            itk_image = itk.imread(ensure_filename_is_str(path))
        return itk_image

    def create_attenuation_image(self, database, energy):
        # convert all materials to mu
        label_to_mu = {}
        mu_handler = g4.GateMaterialMuHandler.GetInstance(
            database, 200 * g4_units.MeV
        )  # max in MeV
        prod_cuts_table = g4.G4ProductionCutsTable.GetProductionCutsTable()
        for i in range(prod_cuts_table.GetTableSize()):
            couple = prod_cuts_table.GetMaterialCutsCouple(i)
            mat_name = str(couple.GetMaterial().GetName())
            label = self.material_to_label_lut[mat_name]
            mu = mu_handler.GetMu(couple, energy / g4_units.MeV)
            label_to_mu[label] = mu

        arr = itk.GetArrayViewFromImage(self.label_image)
        mu_arr = arr.copy().astype("float32")
        for label, mu in label_to_mu.items():
            mu_arr[mu_arr == label] = mu
        itk_mu_img = itk.GetImageFromArray(mu_arr)
        itk_mu_img.CopyInformation(self.itk_image)
        return itk_mu_img

    def create_label_image(self, itk_image=None):
        # read image
        if itk_image is None:
            if self.itk_image is None:
                self.itk_image = self.load_input_image()
            itk_image = self.itk_image

        if self.material_to_label_lut is None:
            self.material_to_label_lut = self.create_material_to_label_lut()

        # sort voxel_materials according to lower bounds
        voxel_materials_sorted = sorted(self.voxel_materials, key=lambda x: x[0])

        # find gaps in the voxels materials intervals
        # upper bounds that are not also lower bounds of the subsequent interval
        bins = [row[0] for row in voxel_materials_sorted]
        # additional bins are those where an upper interval boundary
        # does not correspond to the next lower interval boundary
        additional_bins = set([row[1] for row in voxel_materials_sorted]).difference(
            bins
        )
        bins.extend(additional_bins)
        labels = [self.material_to_label_lut[row[2]] for row in voxel_materials_sorted]
        labels.extend(
            len(additional_bins) * [0]
        )  # additional bins should have label 0,
        # i.e. the volume's standard material
        # np.digitize function requires bins in ascending order -> sort
        bins_sorted = []
        labels_sorted = [0]  # label 0 for voxel values below lowest interval
        for b, l in sorted(zip(bins, labels), key=lambda pair: pair[0]):
            bins_sorted.append(b)
            labels_sorted.append(l)

        # get numpy array view of input itk image
        input_image = itk.array_view_from_image(itk_image)

        label_image_arr = np.array(labels_sorted, dtype=np.ushort)[
            np.digitize(input_image, bins=bins_sorted)
        ]

        label_image = itk.image_from_array(label_image_arr)
        label_image.CopyInformation(itk_image)
        return label_image

    def create_image_parametrisation(self, label_image=None):
        if label_image is None:
            if self.label_image is None:
                self.label_image = self.create_label_image()
            label_image = self.label_image
        # initialize parametrisation
        g4_voxel_param = g4.GateImageNestedParameterisation()

        # send image to cpp size
        update_image_py_to_cpp(label_image, g4_voxel_param.cpp_edep_image, True)
        g4_voxel_param.initialize_image()
        g4_voxel_param.initialize_material(list(self.material_to_label_lut.keys()))

        return g4_voxel_param

    def update_label_image(self, label_image):
        """Needed for dynamic image parametrisation."""
        # send image to cpp size
        update_image_py_to_cpp(label_image, self.g4_voxel_param.cpp_edep_image, True)
        self.g4_voxel_param.initialize_image()

    def save_label_image(self, path=None):
        # dump label image ?
        if path is None:
            if self.volume_manager is None:
                fatal(
                    f"Cannot save label image of ImageVolume {self.name}. "
                    f"Either provide a path or add the volume to the simulation. "
                )
            root, ext = os.path.splitext(self.dump_label_image)
            # path = (
            #    self.volume_manager.simulation.get_output_path()
            #    / f"label_to_material_lut_{self.name}.json"
            # )
            path = root + ".json"
        if self.label_image is None:
            self.create_label_image()

        self.label_image.SetOrigin(self.itk_image.GetOrigin())  # set origin as in input
        # FIXME: should write image into output dir
        write_itk_image(self.label_image, str(self.dump_label_image))
        with open(path, "w") as f:
            json.dump(self.material_to_label_lut, f)

        # re-compute image origin such that it is centered at 0
        self.label_image.SetOrigin(
            -(self.size_pix * self.spacing) / 2.0 + self.spacing / 2.0
        )

    def create_density_image(self):
        return create_density_img(self, self.volume_manager.material_database)

    def create_changers(self):
        # get the changers from the mother classes and append those specific to the ImageVolume class
        changers = super().create_changers()
        counter = 0
        for dp in self.dynamic_params.values():
            if dp["extra_params"]["auto_changer"] is True:
                if "image" in dp:
                    # create a LUT of image parametrisations
                    label_image = {}
                    for path_to_image in set(dp["image"]):
                        itk_image = self.load_input_image(path_to_image)
                        label_image[path_to_image] = self.create_label_image(itk_image)
                    new_changer = VolumeImageChanger(
                        name=f"{self.name}_volume_image_changer_{len(changers)}",
                        attached_to=self,
                        simulation=self.simulation,
                        images=dp["image"],
                        label_image=label_image,
                    )
                    changers.append(new_changer)
                    counter += 1
            else:
                warning(
                    f"You need to manually create a changer for dynamic parametrisation {dp} "
                    f"of volume '{self.name}'."
                )
        if counter > 1:
            warning(
                f"You have provided multiple dynamic image parametrisation (4D image) "
                f"in the {type(self).__name__} named {self.name}. "
                f"Consider verifying if this is intentional. "
            )
        return changers

    def write_material_database(self, material_filename):
        # get all the materials names
        materials = [m[2] for m in self.voxel_materials]
        # Maintaining order while keeping unique
        unique_materials = list(dict.fromkeys(materials))
        write_material_database(self.simulation, unique_materials, material_filename)

    def write_label_to_material(self, labels_filename):
        with open(labels_filename, "w") as outfile:
            dump_json(self.voxel_materials, outfile, indent=4)

    def read_label_to_material(self, labels_filename):
        with open(labels_filename, "r") as infile:
            labels = json.load(infile)
        self.voxel_materials = labels


class ParallelWorldVolume(NodeMixin):
    def __init__(self, name, volume_manager):
        # VolumeBase.__init__(self, name)
        # NodeMixin.__init__(self)
        super().__init__()
        self.name = name
        self.volume_manager = volume_manager
        # the volume manager is guaranteed to have a volume_tree_root because it is created in __init__
        self.parent = self.volume_manager.volume_tree_root

        # it is attached to the parallel world engine, instead of the volume engine as other volumes
        self.parallel_world_engine = None

        self.g4_world_phys_vol = None
        self.g4_world_log_vol = None

    def release_g4_references(self):
        self.g4_world_phys_vol = None
        self.g4_world_log_vol = None

    def close(self):
        self.release_g4_references()
        self.parallel_world_engine = None

    def __getstate__(self):
        return_dict = self.__dict__
        return_dict["g4_world_phys_vol"] = None
        return_dict["g4_world_log_vol"] = None
        return_dict["parallel_world_engine"] = None
        return return_dict

    @requires_fatal("parallel_world_engine")
    def construct(self):
        # get the physical volume through the parallel world engine
        # do not construct it
        self.g4_world_phys_vol = self.parallel_world_engine.GetWorld()
        self.g4_world_log_vol = self.g4_world_phys_vol.GetLogicalVolume()

    # need this dummy method because the parent attribute should not be updated
    def _update_node(self):
        pass


# inherit from NodeMixin turn the class into a tree node
class VolumeTreeRoot(NodeMixin):
    """Small class to provide a root for the volume tree."""

    def __init__(self, volume_manager) -> None:
        super().__init__()
        self.volume_manager = volume_manager
        self.name = "volume_tree_root"
        self.parent = None  # None means this is a tree root

    def __getstate__(self):
        return_dict = self.__dict__
        return_dict["volume_engine"] = None
        return return_dict


# The following lines make sure that all classes which
# inherit from the GateObject base class are processed upon importing opengate.
# In this way, all properties corresponding to the class's user_info dictionary
# will be created.
# This ensures, e.g., that auto-completion in interactive python consoles
# and code editors suggests the properties.
process_cls(VolumeBase)
process_cls(BooleanVolume)
process_cls(RepeatableVolume)
process_cls(BoxVolume)
process_cls(HexagonVolume)
process_cls(ConsVolume)
process_cls(PolyhedraVolume)
process_cls(SphereVolume)
process_cls(EllipsoidVolume)
process_cls(TrapVolume)
process_cls(TrdVolume)
process_cls(TubsVolume)
process_cls(RepeatParametrisedVolume)
process_cls(ImageVolume)
process_cls(TesselatedVolume)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import numpy as np
import matplotlib.pyplot as plt
import os
import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
import opengate.contrib.phantoms.nemaiec as gate_iec
import gatetools.phsp as phsp
from opengate.tests import utility


class GANTest:
    def __init__(self, spheres_activity_ratio, spheres_centers, spheres_radius, rs):
        # will store all conditional info (position, direction)
        # (not needed, only for test)
        self.all_cond = None
        self.spheres_activity_ratio = spheres_activity_ratio
        self.spheres_centers = spheres_centers
        self.spheres_radius = spheres_radius
        self.rs = rs

    def generate_condition(self, n):
        n_samples = gate_iec.get_n_samples_from_ratio(n, self.spheres_activity_ratio)
        cond = gate_iec.generate_pos_dir_spheres(
            self.spheres_centers,
            self.spheres_radius,
            n_samples,
            shuffle=True,
            rs=self.rs,
        )

        if self.all_cond is None:
            self.all_cond = cond
        else:
            self.all_cond = np.column_stack((self.all_cond, cond))

        return cond


def create_simulation(sim, paths, colli="lehr", version=""):
    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3

    # main parameters
    sim.check_volumes_overlap = True
    sim.random_seed = 4123456
    # ac = 1e6 * BqmL
    ac = 3e3 * BqmL / sim.number_of_threads
    sim.visu = False
    # sim.running_verbose_level = gate.EVENT
    # sim.g4_verbose = True
    sim.output_dir = paths.output

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # cylinder of the phase space, for visualisation only
    """cyl = sim.add_volume('Sphere', 'phase_space_cylinder')
    cyl.rmin = 210 * mm
    cyl.rmax = 211 * mm
    cyl.color = [1, 1, 1, 1]
    cyl.material = 'G4_AIR'"""

    # test phase space to check with reference
    phase_space_sphere = sim.add_volume("Sphere", "phase_space_sphere")
    phase_space_sphere.rmin = 212 * mm
    phase_space_sphere.rmax = 213 * mm
    phase_space_sphere.color = [1, 1, 1, 1]
    phase_space_sphere.material = "G4_AIR"

    # spect head
    distance = 30 * cm
    psd = 6.11 * cm
    p = [0, 0, -(distance + psd)]
    spect1, colli, crystal = gate_spect.add_spect_head(
        sim, "spect1", collimator_type=colli, debug=sim.visu
    )
    spect1.translation, spect1.rotation = gate.geometry.utility.get_transform_orbiting(
        p, "x", 180
    )

    # physic list
    sim.physics_manager.set_production_cut("world", "all", 1 * mm)

    # activity parameters
    spheres_diam = [10, 13, 17, 22, 28, 37]
    spheres_activity_concentration = [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac]

    # initialisation for conditional
    spheres_radius = [x / 2.0 for x in spheres_diam]
    (
        spheres_centers,
        spheres_volumes,
    ) = gate_iec.get_default_sphere_centers_and_volumes_old()
    spheres_activity_ratio = []
    spheres_activity = []
    for diam, ac, volume, center in zip(
        spheres_diam, spheres_activity_concentration, spheres_volumes, spheres_centers
    ):
        activity = ac * volume
        print(
            f"Sphere {diam}: {str(center):<30} {volume / cm3:7.3f} cm3 "
            f"{activity / Bq:7.0f} Bq  {ac / BqmL:7.1f} BqmL"
        )
        spheres_activity.append(activity)

    total_activity = sum(spheres_activity)
    print(f"Total activity {total_activity / Bq:.0f} Bq")
    for activity in spheres_activity:
        spheres_activity_ratio.append(activity / total_activity)
    print("Activity ratio ", spheres_activity_ratio, sum(spheres_activity_ratio))

    # unique (reproducible) random generator
    rs = gate.utility.get_rnd_seed(123456)

    # GAN source
    gsource = sim.add_source("GANSource", "gaga")
    gsource.particle = "gamma"
    # no phantom, we consider attached to the world at origin
    # gsource.attached_to = f'{iec_phantom.name}_interior'
    gsource.activity = total_activity
    gsource.pth_filename = paths.gate / "pth2" / "test001_GP_0GP_10_50000.pth"
    gsource.position_keys = ["PrePosition_X", "PrePosition_Y", "PrePosition_Z"]
    gsource.backward_distance = 5 * cm
    gsource.direction_keys = ["PreDirection_X", "PreDirection_Y", "PreDirection_Z"]
    gsource.energy_key = "KineticEnergy"
    # gsource.energy_threshold = 0.001 * keV
    gsource.energy_min_threshold = 10 * keV
    # gsource.skip_policy = "SkipEvents"
    # SkipEvents is a bit faster than Energy zero,
    # but it changes the nb of events,so force ZeroEnergy
    gsource.skip_policy = "ZeroEnergy"
    gsource.weight_key = None
    gsource.time_key = "TimeFromBeginOfEvent"
    gsource.relative_timing = True
    gsource.batch_size = 5e4
    gsource.verbose_generator = True
    gsource.gpu_mode = (
        utility.get_gpu_mode_for_tests()
    )  # should be "auto" but "cpu" for macOS github actions to avoid mps errors

    # GANSourceConditionalGenerator manages the conditional GAN
    # GANTest manages the generation of the conditions, we use a class here to store the total
    # list of conditions (only needed for the test)
    condition_generator = GANTest(
        spheres_activity_ratio, spheres_centers, spheres_radius, rs
    )
    gen = gate.sources.gansources.GANSourceConditionalGenerator(
        gsource, condition_generator.generate_condition
    )
    gsource.generator = gen

    # it is possible to use acceptance angle. Not done here to check exiting phsp
    # gsource.direction.acceptance_angle.volumes = [spect1.name]
    # gsource.direction.acceptance_angle.intersection_flag = True

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.output_filename = f"test038_gan_stats{version}.txt"

    # add default digitizer (it is easy to change parameters if needed)
    gate_spect.add_simplified_digitizer_tc99m_OLD(
        sim, "spect1_crystal", "test038_gan_proj.mhd"
    )
    # gate_spect.add_ge_nm670_spect_simplified_digitizer(sim, 'spect2_crystal', paths.output / 'test033_proj_2.mhd')
    singles_actor = sim.actor_manager.get_actor(f"Singles_spect1_crystal")
    singles_actor.output_filename = f"test038_gan_singles{version}.root"

    # motion of the spect, create also the run time interval
    """heads = [spect1]  # [spect1, spect2]

    # create a list of run (total = 1 second)
    n = 1
    sim.run_timing_intervals = gate.range_timing(0, 1 * sec, n)

    for head in heads:
        motion = sim.add_actor('MotionVolumeActor', f'Move_{head.name}')
        motion.mother = head.name
        motion.translations, motion.rotations = \
            gate.volume_orbiting_transform('x', 0, 180, n, head.translation, head.rotation)
        motion.priority = 5"""

    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp_actor.attached_to = phase_space_sphere.name
    phsp_actor.attributes = [
        "KineticEnergy",
        "PrePosition",
        "PreDirection",
        "GlobalTime",
        "EventPosition",
        "EventDirection",
        "EventKineticEnergy",
    ]
    phsp_actor.output_filename = f"test038_gan_phsp{version}.root"

    return condition_generator


def analyze_results(sim, paths, all_cond):
    phsp_actor = sim.get_actor("phsp")
    print(phsp_actor)

    # print stats
    print()
    gate.exception.warning(f"Check stats")
    s = sim.get_source_user_info("gaga")
    print(f"Source, nb of skipped particles (absorbed) : {s.total_skipped_events}")
    print(f"Source, nb of zeros   particles (absorbed) : {s.total_zero_events}")

    stats = sim.get_actor("Stats")
    print(stats)
    stats.counts.events += s.total_skipped_events
    stats_ref = utility.read_stat_file(paths.output_ref / "test038_ref_stats.txt")
    r = (stats_ref.counts.steps - stats.counts.steps) / stats_ref.counts.steps
    print(f"Steps cannot be compared => was {stats.counts.steps}, {r:.2f}%")
    stats.counts.steps = stats_ref.counts.steps
    if s.total_skipped_events > 0:
        print(f"Tracks cannot be compared => was {stats.counts.tracks}")
        stats.counts.tracks = stats_ref.counts.tracks

    stats.counts.runs = 1  # force for MT
    is_ok = utility.assert_stats(stats, stats_ref, 0.10)

    # save conditional for checking with reference cond
    keys = [
        "EventPosition_X",
        "EventPosition_Y",
        "EventPosition_Z",
        "EventDirection_X",
        "EventDirection_Y",
        "EventDirection_Z",
    ]
    phsp.save_npy(paths.output / "test038_gan_phsp_cond.npy", all_cond, keys)

    # ----------------------------------------------------------------------------------------------
    # compare conditional
    # less particle in the ref because conditional data are stored
    # when exit (not absorbed)
    print()
    gate.exception.warning(f"Check conditions (position, direction)")
    root_ref = (
        paths.output_ref / "test038_ref_phsp.root"
    )  # looking the previous generated
    hits1 = uproot.open(root_ref)
    branch = hits1.keys()[0]
    print("Branch name:", branch)
    hits1 = hits1[branch]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")
    root_gan = paths.output / "test038_gan_phsp_cond.npy"
    hits2, hits2_keys, hits2_n = phsp.load(root_gan)
    tols = [10.0] * len(keys)
    tols[keys.index("EventPosition_X")] = 0.3
    tols[keys.index("EventPosition_Y")] = 0.5
    tols[keys.index("EventPosition_Z")] = 0.3
    tols[keys.index("EventDirection_X")] = 0.02
    tols[keys.index("EventDirection_Y")] = 0.02
    tols[keys.index("EventDirection_Z")] = 0.03
    scalings = [1] * len(keys)
    is_ok = (
        utility.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2_keys),
            keys,
            keys,
            tols,
            scalings,
            scalings,
            True,
        )
        and is_ok
    )
    # figure
    img_filename = paths.output / "test038_cond.png"
    plt.suptitle(
        f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img_filename)
    print(f"Figure in {img_filename}")

    # ----------------------------------------------------------------------------------------------
    # compare output phsp
    """
        This is *not* a very good pth for the moment, we set a high tolerance.
    """
    print()
    gate.exception.warning(f"Check output phsp")
    ref_file = paths.output_ref / "test038_ref_phsp.root"
    hc_file = phsp_actor.get_output_path()
    checked_keys = [
        "GlobalTime",
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
    ]
    scalings = [1.0] * len(checked_keys)
    scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
    tols = [10.0] * len(checked_keys)
    tols[checked_keys.index("GlobalTime")] = 0.2
    tols[checked_keys.index("KineticEnergy")] = 0.002
    tols[checked_keys.index("PrePosition_X")] = 7
    tols[checked_keys.index("PrePosition_Y")] = 5.5
    tols[checked_keys.index("PrePosition_Z")] = 5.1
    tols[checked_keys.index("PreDirection_X")] = 0.03
    tols[checked_keys.index("PreDirection_Y")] = 0.02
    tols[checked_keys.index("PreDirection_Z")] = 0.03
    print(scalings, tols)
    is_ok = (
        utility.compare_root3(
            ref_file,
            hc_file,
            "phsp",
            "phsp",
            checked_keys,
            checked_keys,
            tols,
            scalings,
            scalings,
            paths.output / "test038_phsp.png",
        )
        and is_ok
    )

    # ----------------------------------------------------------------------------------------------
    # compare hits
    print()
    gate.exception.warning(f"Check singles -> NOT YET (too low statistics)")

    """ref_file = paths.output / 'test038_ref_singles.root'
    hc_file = singles_actor.output
    checked_keys = ['GlobalTime', 'TotalEnergyDeposit', 'PostPosition_X', 'PostPosition_Y', 'PostPosition_Z']
    scalings = [1.0] * len(checked_keys)
    scalings[checked_keys.index('GlobalTime')] = 1e-9  # time in ns
    tols[checked_keys.index('GlobalTime')] = 0.2
    tols[checked_keys.index('TotalEnergyDeposit')] = 10
    tols[checked_keys.index('PostPosition_X')] = 100
    tols[checked_keys.index('PostPosition_Y')] = 100
    tols[checked_keys.index('PostPosition_Z')] = 100
    print(scalings, tols)
    is_ok = utility.compare_root3(ref_file, hc_file, "Singles_spect1_crystal", "Singles_spect1_crystal",
                              checked_keys, checked_keys, tols, scalings, scalings,
                              paths.output / 'test038_singles.png', hits_tol=100) and is_ok
    """
    # ----------------------------------------------------------------------------------------------

    # this is the end, my friend
    # gate.delete_run_manager_if_needed(sim)
    utility.test_ok(is_ok)
